# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""
Task to test Debian binary packages using piuparts.

This task follows the specification laid out here:
https://freexian-team.pages.debian.net/debusine/reference/tasks.html#piuparts-task
"""

import gzip
import re
import shlex
import shutil
import tarfile
from collections.abc import Collection
from io import BytesIO
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import override

from debian import debfile
from debian.debian_support import Version

import debusine.utils
from debusine.artifacts import PiupartsArtifact
from debusine.artifacts.models import (
    ArtifactCategory,
    DebianBinaryPackage,
    DebianPiuparts,
    DebianSystemTarball,
    DebianUpload,
    WorkRequestResults,
    get_source_package_name,
    get_source_package_version,
)
from debusine.client.models import RelationType
from debusine.tasks import (
    BaseTaskWithExecutor,
    ExtraRepositoryMixin,
    RunCommandTask,
    TaskConfigError,
)
from debusine.tasks.executors import ExecutorImageCategory
from debusine.tasks.inputs import EnvironmentInput, MultiInput, Stage
from debusine.tasks.models import BackendType, PiupartsData, PiupartsDynamicData


class Piuparts(
    ExtraRepositoryMixin[PiupartsData, PiupartsDynamicData],
    RunCommandTask[PiupartsData, PiupartsDynamicData],
    BaseTaskWithExecutor[PiupartsData, PiupartsDynamicData],
):
    """Test Debian binary packages using piuparts."""

    TASK_VERSION = 1

    CAPTURE_OUTPUT_FILENAME = "piuparts.txt"

    base_tgz = EnvironmentInput(
        image_category=ExecutorImageCategory.TARBALL,
        set_backend=False,
        try_variant=False,
        categories=[ArtifactCategory.SYSTEM_TARBALL],
        stage=Stage.PENDING,
    )
    environment = EnvironmentInput()
    input_binary_artifacts = MultiInput(
        field="input.binary_artifacts",
        categories=[ArtifactCategory.BINARY_PACKAGE, ArtifactCategory.UPLOAD],
        stage=Stage.PENDING,
    )

    @override
    def post_init(self) -> None:
        super().post_init()

        # *.deb paths and versions. Set by self.configure_for_execution()
        self._deb_files: list[Path] = []
        self._deb_versions: dict[str, str] | None = None

        # Version of piuparts in the executor.
        # Set by self.configure_for_execution()
        self._piuparts_version: Version | None = None

        # --basetgz=... option for piuparts cmdline.
        # Set by self.configure_for_execution()
        self._base_tar: Path | None = None
        self._base_tar_data: DebianSystemTarball | None = None
        self._scripts_dir: Path | None = None

    @override
    def get_subject(self) -> str:
        package_names = set()
        for binary_artifact in self.input_binary_artifacts.artifacts:
            assert isinstance(
                binary_artifact.data,
                (DebianBinaryPackage, DebianUpload),
            )
            package_names.add(get_source_package_name(binary_artifact.data))

        package_names_sorted = sorted(package_names)
        return " ".join(package_names_sorted)

    @override
    def get_configuration_context(self) -> str | None:
        assert isinstance(self.base_tgz.data, DebianSystemTarball)
        return self.base_tgz.data.codename

    @override
    def compute_dynamic_data(self) -> PiupartsDynamicData:
        """
        Resolve artifact lookups for this task.

        :subject: space-separated list of source package names from
          ``input.binary_artifact``
        :parameter_summary: same as subject with the version (if one package)
        :runtime_context: ``codename:build_architecture`` where ``codename``
          is from ``base_tgz``
        :configuration_context: ``codename`` from ``base_tgz``
        """
        base_tgz = self.base_tgz
        assert isinstance(base_tgz.data, DebianSystemTarball)

        source_package_names = set()
        source_versions = set()
        for binary_artifact in self.input_binary_artifacts.artifacts:
            assert isinstance(
                binary_artifact.data,
                (DebianBinaryPackage, DebianUpload),
            )
            source_package_names.add(
                get_source_package_name(binary_artifact.data)
            )
            source_versions.add(
                get_source_package_version(binary_artifact.data)
            )

        runtime_context = f"{base_tgz.data.codename}:{self.build_architecture}"

        subject = self.get_subject()
        if len(source_package_names) == 1 and len(source_versions) == 1:
            (source_version,) = source_versions
            (source_package_name,) = source_package_names
            parameter_summary = f"{source_package_name}_{source_version}"
        else:
            parameter_summary = subject

        # TODO: It would probably be better to get the architecture of the
        # input binary artifacts and use that.
        parameter_summary += f" ({self.build_architecture})"

        return PiupartsDynamicData(
            environment_id=self.environment.artifact_id,
            extra_repositories=self.extra_repositories,
            input_binary_artifacts_ids=self.input_binary_artifacts.resolved_ids(),
            base_tgz_id=base_tgz.artifact_id,
            subject=subject,
            parameter_summary=parameter_summary,
            runtime_context=runtime_context,
            configuration_context=self.get_configuration_context(),
        )

    @override
    def fetch_input(self, destination: Path) -> bool:
        """Populate work directory with user-specified binary artifact(s)."""
        assert self.dynamic_data

        for artifact_id in self.dynamic_data.input_binary_artifacts_ids:
            self.fetch_artifact(artifact_id, destination)
        return True

    @override
    def configure_for_execution(self, download_directory: Path) -> bool:
        r"""
        Find the .deb files for piuparts.

        Prepare executor, install "piuparts" in it and prepare
        debian:system-image (e.g. delete /dev/\* files).

        Set self._deb_files to the relevant files.

        :param download_directory: where to find the \*.deb files (downloaded
           via fetch_input) and where to download the chroot of
           debian:system-image (for piuparts --basetgz).
        :return: True if valid files were found
        """
        # Find the files to test or early exit if no files
        self._deb_files = debusine.utils.find_files_suffixes(
            download_directory, [".deb"]
        )
        # Ensure we've got >=1 .deb file(s)
        # Note: This could be moved into a more generic function shared across
        # tasks.
        if len(self._deb_files) == 0:
            list_of_files = sorted(map(str, download_directory.iterdir()))
            self.append_to_log_file(
                "configure_for_execution.log",
                [
                    f"There must be at least one *.deb file. "
                    f"Current files: {list_of_files}"
                ],
            )
            return False

        deb_versions: dict[str, str] = {}
        for file in self._deb_files:
            pkg = debfile.DebFile(file)
            try:
                control = pkg.control.debcontrol()
                deb_versions[control["Package"]] = control["Version"]
            finally:
                pkg.close()
        self._deb_versions = deb_versions

        # Prepare executor
        self._prepare_executor_instance()

        if self.executor_instance is None:
            raise AssertionError("self.executor_instance cannot be None")

        # Prepare executor_instance to run piuparts
        self.run_executor_command(
            ["apt-get", "update"],
            log_filename="install.log",
            run_as_root=True,
            check=True,
        )

        self.run_executor_command(
            ["apt-get", "--yes", "install", "piuparts"],
            log_filename="install.log",
            run_as_root=True,
            check=True,
        )

        self._check_piuparts_version()

        self._prepare_base_tgz(download_directory)
        self._scripts_dir = download_directory / "scripts"
        self._scripts_dir.mkdir()
        self._prepare_scripts(self._scripts_dir)

        return True

    def _check_piuparts_version(self) -> None:
        """Check that the environment contains a supported piuparts version."""
        # We are root, older piuparts will work
        if self.backend in (BackendType.INCUS_VM, BackendType.QEMU):
            return

        # We are inside a container, and can't create device nodes (#638)
        self._piuparts_version = self.get_package_version("piuparts")
        min_version = Version("1.5")

        if self._piuparts_version < min_version:
            raise TaskConfigError(
                f"The environment contains a version of piuparts "
                f"({self._piuparts_version}) that is too old. "
                f"Container-based backends require piuparts >= "
                f"{min_version}. We recommend a >= trixie environment "
                f"with an older base_tgz."
            )

    def _prepare_base_tgz(self, download_directory: Path) -> None:
        """
        Download the base image for piuparts. Remove /dev/* if needed.

        Set self._base_tar to the file that is used by _cmdline.
        """
        if self.executor_instance is None:
            raise AssertionError("self.executor_instance cannot be None")
        assert self.dynamic_data

        (base_tar_dir := download_directory / "base_tar").mkdir()

        base_tgz_response = self.fetch_artifact(
            self.dynamic_data.base_tgz_id, base_tar_dir
        )
        self._base_tar_data = DebianSystemTarball.model_validate(
            base_tgz_response.data
        )
        self._base_tar = base_tar_dir / self._base_tar_data.filename

        # Is the base_tgz set up for systemd-resolved?
        resolvconf_symlink = False
        # Does /var/lib/dpkg/available exist?
        dpkg_available = False
        with tarfile.open(self._base_tar) as tar:
            first_member = tar.getmembers()[0]
            # How are the members prefixed? Typically ./
            m = re.search(r"^(?:\.?/)?", first_member.name)
            assert m  # This regex would match an empty string
            prefix = m.group(0)

            try:
                if tar.getmember(
                    f"{prefix}etc/resolv.conf"
                ).issym():  # pragma: no cover
                    resolvconf_symlink = True
            except KeyError:
                pass

            try:
                if tar.getmember(
                    f"{prefix}var/lib/dpkg/available"
                ):  # pragma: no cover
                    dpkg_available = True
            except KeyError:
                pass

        if (
            not base_tgz_response.data["with_dev"]
            and not resolvconf_symlink
            and dpkg_available
        ):
            # No processing needs to be done, the image can stay as
            # .tar.xz (or the format retrieved)
            return

        # The tar will be fixed up and self._base_tar pointed at the new file

        filter_dirs: list[Path] = []
        filter_files: list[Path] = []
        create_empty_files: list[Path] = []
        if base_tgz_response.data["with_dev"]:
            filter_dirs.append(Path("/dev"))
        if resolvconf_symlink:
            # Replace a broken symlink with an empty file
            filter_files.append(Path("/etc/resolv.conf"))
            create_empty_files.append(Path("/etc/resolv.conf"))
        if not dpkg_available:
            create_empty_files.append(Path("/var/lib/dpkg/available"))

        processed_tar_gz = self._filter_tar(
            self._base_tar,
            filter_dirs=filter_dirs,
            filter_files=filter_files,
            create_empty_files=create_empty_files,
        )

        # Use the processed image
        self._base_tar = processed_tar_gz

    def _filter_tar(
        self,
        input_: Path,
        filter_dirs: Collection[Path] = (),
        filter_files: Collection[Path] = (),
        create_empty_files: Collection[Path] = (),
    ) -> Path:
        """Process a tarball, filtering contents and creating empty files."""
        processed_tar_name = input_.stem.rsplit(".", 1)[0] + ".tar"
        output = input_.with_name(processed_tar_name)
        if input_ == output:
            raise AssertionError(
                "We generated a filtered filename that matched the input."
            )

        with (
            tarfile.open(input_) as input_tar,
            tarfile.open(output, mode="w") as output_tar,
        ):
            for member in input_tar:
                path = Path("/" + member.name.lstrip('./'))
                if any(path.is_relative_to(dir_) for dir_ in filter_dirs):
                    continue
                if path in filter_files:
                    continue

                if member.isfile():
                    output_tar.addfile(
                        member, input_tar.extractfile(member.name)
                    )
                else:
                    output_tar.addfile(member)

            for path in create_empty_files:
                for parent in reversed(path.parents):
                    if parent == Path("/"):
                        continue
                    try:
                        if output_tar.getmember(f".{parent}").isdir():
                            continue
                        raise AssertionError(
                            f"{parent} exists in the tarball, but not a "
                            "directory."
                        )
                    except KeyError:
                        tarinfo = tarfile.TarInfo(name=f".{parent}")
                        tarinfo.mode = 0o755
                        tarinfo.type = tarfile.DIRTYPE
                        output_tar.addfile(tarinfo)
                tarinfo = tarfile.TarInfo(name=f".{path}")
                tarinfo.mode = 0o644
                tarinfo.size = 0
                output_tar.addfile(tarinfo, BytesIO(b""))

        # Compress the processed tarball for compatibility with piuparts < 1.3
        output_tar_gz = output.with_suffix(".tar.gz")
        with (
            open(output, "rb") as uncompressed,
            gzip.open(output_tar_gz, "wb") as compressed,
        ):
            shutil.copyfileobj(uncompressed, compressed)

        output.unlink()

        return output_tar_gz

    def _prepare_scripts(self, script_directory: Path) -> None:
        """Prepare scripts for piuparts."""
        if not self._base_tar_data:
            raise AssertionError(
                "self._base_tar_data not set, call _prepare_base_tgz first."
            )

        bad_permissions_script = (
            script_directory / "pre_remove_50_find_bad_permissions"
        )
        bad_permissions_script.write_text(
            "#!/bin/sh\n"
            "echo 'Skipping to avoid https://bugs.debian.org/1125784'\n"
        )
        bad_permissions_script.chmod(0o755)

        install_passwd_script = script_directory / "post_setup_00_passwd"
        install_passwd_script.write_text(
            "#!/bin/sh\n"
            "echo 'Installing passwd for https://bugs.debian.org/1125789'\n"
            "if [ ! -e /usr/sbin/useradd ]; then\n"
            "    apt-get -y install passwd\n"
            "fi\n"
        )
        install_passwd_script.chmod(0o755)

        codename = self._base_tar_data.codename
        if self.data.extra_repositories:
            extra_repo_script = (
                script_directory / "post_chroot_unpack_debusine_extra_repos"
            )
            with (
                TemporaryDirectory(
                    prefix="debusine-piuparts-"
                ) as extra_repositories_dir,
                extra_repo_script.open("w") as f,
            ):
                extra_repositories_path = Path(extra_repositories_dir)
                self.write_extra_repository_config(
                    codename=codename, destination=extra_repositories_path
                )

                f.write("#!/bin/sh\nset -eux\n")
                if self.extra_repository_keys:
                    f.write("mkdir -p /etc/apt/keyrings\n")
                for key in self.extra_repository_keys:
                    f.write(
                        f"printf %s {shlex.quote(key.read_text())}"
                        f" > /etc/apt/keyrings/{key.name}\n"
                    )
                for source in (
                    self.extra_repository_sources + self.extra_repository_lists
                ):
                    f.write(
                        f"printf %s {shlex.quote(source.read_text())}"
                        f" > /etc/apt/sources.list.d/{source.name}\n"
                    )
                f.write("echo DONE\n")
            extra_repo_script.chmod(0o755)

    @override
    def _cmdline(self) -> list[str]:
        """Build full piuparts command line."""
        cmdline = ["piuparts"]

        # Use --keep-sources-list (not --distribution=dist-variant /
        # --mirror / --extra-repo): we feed pre-generated tarballs and
        # we don't want piuparts to overwrite the APT configuration.
        # Also don't set --distribution/-d: we don't want to
        # work-around dated distros.conf aliasing (stable/testing/...)
        # and variants (-backports,-updates,/updates,-security/...).
        # Only use -d in the future if we need to enable --scriptsdir
        # and need to set $PIUPARTS_DISTRIBUTION.
        cmdline.append("--keep-sources-list")

        # Common options used in Debian piuparts services/CI
        cmdline.append("--allow-database")
        cmdline.append("--warn-on-leftovers-after-purge")

        # Use pre-generated tarballs (don't debootstrap on the fly)
        cmdline.append(f"--basetgz={self._base_tar}")

        cmdline.append("--scriptsdir=/etc/piuparts/scripts")

        # Configure APT sources, etc.
        cmdline.append(f"--scriptsdir={self._scripts_dir}")

        # Test the binary .deb files
        cmdline.extend(map(str, self._deb_files))

        return cmdline

    @override
    @staticmethod
    def _cmdline_as_root() -> bool:
        """Piuparts must run as root."""
        return True

    @override
    def check_directory_for_consistency_errors(
        self, build_directory: Path
    ) -> list[str]:
        """Return list of errors."""
        if build_directory / self.CAPTURE_OUTPUT_FILENAME not in set(
            build_directory.iterdir()
        ):
            return [f"{self.CAPTURE_OUTPUT_FILENAME} not in {build_directory}"]

        return []

    @override
    def upload_artifacts(
        self,
        execute_directory: Path,
        *,
        execution_result: WorkRequestResults,  # noqa: U100
    ) -> None:
        """Upload a PiupartsArtifact with the files, data, and relationships."""
        assert self.debusine is not None
        assert self._deb_versions is not None
        assert self._piuparts_version is not None
        assert self._base_tar_data is not None

        base_tar_data = self._base_tar_data
        piuparts_artifact = PiupartsArtifact.create(
            piuparts_output=execute_directory / self.CAPTURE_OUTPUT_FILENAME,
            data=DebianPiuparts(
                piuparts_version=str(self._piuparts_version),
                binary_packages=self._deb_versions,
                # It may seem as though this should be the architecture of
                # the binary packages instead (for example, we might be
                # testing `Architecture: all` binary packages on amd64).
                # However, the results of running `piuparts` inherently
                # depend on the build architecture in other ways even if the
                # input artifacts all happen to be architecture-independent.
                architecture=self.data.build_architecture,
                distribution=f"{base_tar_data.vendor}:{base_tar_data.codename}",
            ),
        )

        uploaded_piuparts_artifact = self.debusine.upload_artifact(
            piuparts_artifact,
            workspace=self.workspace_name,
            work_request=self.work_request_id,
        )

        for source_artifact_id in self._source_artifacts_ids:
            self.debusine.relation_create(
                uploaded_piuparts_artifact.id,
                source_artifact_id,
                RelationType.RELATES_TO,
            )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for the view tests utilities."""

import lxml
import lxml.objectify
from django.urls import reverse

from debusine.test.django import TestCase
from debusine.web.views.tests.utils import ViewTestMixin


class TestWebUtils(ViewTestMixin, TestCase):
    """Tests for view test utility code."""

    def test_workspace_list_not_found(self) -> None:
        """Test that workspace_list_table_rows errors when not found."""
        default_workspace = self.playground.get_default_workspace()
        default_workspace.public = False
        default_workspace.save()

        response = self.client.get(reverse("homepage:homepage"))
        tree = self.assertResponseHTML(response)
        with self.assertRaisesRegex(
            AssertionError, r"page has no workspace list table"
        ):
            self.workspace_list_table_rows(tree)

    def test_assert_html_contents_equivalent(self) -> None:
        """Test that assertHTMLContentsEquivalent works as expected."""
        node = lxml.objectify.fromstring(
            '<span> some  na\n\n (<abbr title="Ext">E</abbr>)\n\n</span>\n     '
        )

        self.assertHTMLContentsEquivalent(
            node, '<span>some    na (<abbr title="Ext">E</abbr>)</span>'
        )

        with self.assertRaises(AssertionError):
            self.assertHTMLContentsEquivalent(node, "<span>Sample</span>")

        with self.assertRaises(AssertionError):
            self.assertHTMLContentsEquivalent(
                node, '<span>so me    na (<abbr title="Ext">E</abbr>)</span>'
            )

    def test_assert_enforces_permission_target_name(self) -> None:
        """Test assertEnforcesPermission target_name resolution."""
        artifact, _ = self.playground.create_artifact()

        self.assertEnforcesPermission(
            artifact.can_display,
            artifact.get_absolute_url(),
            "get_context_data",
        )
        self.assertEnforcesPermission(
            artifact.can_display,
            artifact.get_absolute_url(),
            "debusine.web.views.artifacts.ArtifactDetailView.get_context_data",
        )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Unit tests for the task to copy items into target collections."""

from collections.abc import Sequence
from pathlib import Path
from textwrap import dedent
from typing import Any, ClassVar, override

from django.conf import settings

from debusine.artifacts.models import (
    ArtifactCategory,
    BareDataCategory,
    CollectionCategory,
)
from debusine.client.models import LookupChildType
from debusine.db.models import (
    Artifact,
    Collection,
    CollectionItem,
    FileStore,
    Scope,
    WorkRequest,
    Workspace,
)
from debusine.db.playground import scenarios
from debusine.db.tasks.models import (
    DBInputCollectionSingle,
    DBInputLookupResult,
)
from debusine.server.tasks import CopyCollectionItems
from debusine.server.tasks.copy_collection_items import CannotCopy
from debusine.server.tasks.models import (
    CopyCollectionItemsCopies,
    CopyCollectionItemsData,
)
from debusine.server.tasks.tests.helpers import ServerTaskTestCase
from debusine.tasks.models import LookupMultiple


class CopyCollectionItemsTests(ServerTaskTestCase[CopyCollectionItems]):
    """Tests for :py:class:`CopyCollectionItems`."""

    scenario = scenarios.DefaultContext()

    source_scope: ClassVar[Scope]
    source_workspace: ClassVar[Workspace]
    source_build_logs: ClassVar[Collection]
    target_build_logs: ClassVar[Collection]
    target_build_logs_input: ClassVar[DBInputCollectionSingle]

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        cls.source_scope = cls.playground.get_or_create_scope(
            name="source", label="Source"
        )
        cls.source_workspace = cls.playground.create_workspace(
            scope=cls.source_scope, name="source", public=True
        )
        cls.source_build_logs = cls.playground.create_singleton_collection(
            CollectionCategory.PACKAGE_BUILD_LOGS,
            workspace=cls.source_workspace,
        )
        cls.target_build_logs = cls.playground.create_singleton_collection(
            CollectionCategory.PACKAGE_BUILD_LOGS,
        )
        cls.target_build_logs_input = DBInputCollectionSingle.from_collection(
            cls.target_build_logs,
            lookup=f"{cls.target_build_logs.pk}@collections",
            label="copies[0].target_collection",
        )

    @override
    def setUp(self) -> None:
        super().setUp()
        collection_category = CollectionCategory.PACKAGE_BUILD_LOGS
        self.task_data = {
            "copies": [
                CopyCollectionItemsCopies(
                    source_items={
                        "collection": f"_@{collection_category}",
                        "child_type": LookupChildType.BARE,
                    },
                    target_collection=self.target_build_logs.id,
                )
            ]
        }

    def create_build_log_artifact(
        self,
        work_request: WorkRequest | None = None,
        build_log_contents: bytes | None = None,
        skip_add_files_in_store: bool = False,
        extra_data: dict[str, Any] | None = None,
    ) -> Artifact:
        """Create an artifact with a build log."""
        if work_request is None:
            work_request = self.playground.create_worker_task(
                workspace=self.source_workspace
            )
        artifact = self.playground.create_build_log_artifact(
            source="hello",
            version="1.0-1",
            build_arch="amd64",
            workspace=self.source_workspace,
            work_request=work_request,
            contents=build_log_contents,
            skip_add_files_in_store=skip_add_files_in_store,
        )
        if extra_data is not None:
            artifact.data |= extra_data
            artifact.save()
        return artifact

    def create_build_log(
        self,
        work_request: WorkRequest | None = None,
        build_log_contents: bytes | None = None,
        skip_add_files_in_store: bool = False,
        extra_data: dict[str, Any] | None = None,
    ) -> CollectionItem:
        """Create an artifact collection item with a build log."""
        artifact = self.create_build_log_artifact(
            work_request=work_request,
            build_log_contents=build_log_contents,
            skip_add_files_in_store=skip_add_files_in_store,
            extra_data=extra_data,
        )
        assert artifact.created_by_work_request is not None
        artifact_item_data = {
            "work_request_id": artifact.created_by_work_request.id,
            "vendor": "debian",
            "codename": "bookworm",
            "architecture": "amd64",
        }
        return self.source_build_logs.manager.add_artifact(
            artifact, user=self.scenario.user, variables=artifact_item_data
        )

    def create_build_log_placeholder(self) -> CollectionItem:
        return self.playground.create_bare_data_item(
            self.source_build_logs,
            "debian_bookworm_amd64_hello_1.0-1_1",
            category=BareDataCategory.PACKAGE_BUILD_LOG,
            data={
                "work_request_id": 1,
                "vendor": "debian",
                "codename": "bookworm",
                "architecture": "amd64",
                "srcpkg_name": "hello",
                "srcpkg_version": "1.0-1",
            },
        )

    def prepare_copy_item_test(
        self, lookup: dict[str, Any]
    ) -> Sequence[DBInputLookupResult]:
        """
        Set up a task and lookup results to test ``copy_item``.

        :returns: the :py:class:`DBInputLookupResult` with the source items
          to copy
        """
        source_items = LookupMultiple.model_validate(
            {
                "collection": f"_@{CollectionCategory.PACKAGE_BUILD_LOGS}",
                **lookup,
            }
        )
        self.task_data = {
            "copies": [
                CopyCollectionItemsCopies(
                    source_items=source_items,
                    target_collection=self.target_build_logs.id,
                )
            ]
        }
        results = self.work_request.lookup_multiple(
            source_items,
            workspace=self.source_workspace,
            expect_type=LookupChildType.ANY,
        )
        return [
            DBInputLookupResult.from_lookup_result(
                result, label="copies[0].source_items"
            )
            for result in results
        ]

    def test_compute_scheduler_tags(self) -> None:
        self.assertComputesSchedulerTags()

    def test_copy_item_bare(self) -> None:
        """Copy a bare data item."""
        bare_item = self.create_build_log_placeholder()
        source_items = self.prepare_copy_item_test(
            {"child_type": LookupChildType.BARE}
        )
        self.assertEqual(len(source_items), 1)

        self.task.copy_item(
            source_items[0],
            self.target_build_logs_input,
            user=self.scenario.user,
        )

        self.assertTrue(
            self.target_build_logs.child_items.filter(
                name="debian_bookworm_amd64_hello_1.0-1_1",
                child_type=CollectionItem.Types.BARE,
                category=BareDataCategory.PACKAGE_BUILD_LOG,
                collection__isnull=True,
                artifact__isnull=True,
                data=bare_item.data,
                removed_at__isnull=True,
            ).exists()
        )

    def test_copy_item_bare_needs_unembargo(self) -> None:
        """Copying bare data from private to public needs unembargo=True."""
        self.create_build_log_placeholder()
        source_items = self.prepare_copy_item_test(
            {"child_type": LookupChildType.BARE}
        )
        self.assertEqual(len(source_items), 1)
        self.source_workspace.public = False
        self.source_workspace.save()

        with self.assertRaisesRegex(
            CannotCopy,
            f"Copying from {self.source_workspace}"
            f" to {self.scenario.workspace} "
            f"requires unembargo=True",
        ):
            self.task.copy_item(
                source_items[0],
                self.target_build_logs_input,
                user=self.scenario.user,
            )

    def test_copy_item_bare_with_unembargo(self) -> None:
        """Copying bare data from private to public can unembargo it."""
        bare_item = self.create_build_log_placeholder()
        source_items = self.prepare_copy_item_test(
            {"child_type": LookupChildType.BARE}
        )
        self.assertEqual(len(source_items), 1)
        self.source_workspace.public = False
        self.source_workspace.save()

        self.task.copy_item(
            source_items[0],
            self.target_build_logs_input,
            user=self.scenario.user,
            unembargo=True,
        )

        self.assertTrue(
            self.target_build_logs.child_items.filter(
                name="debian_bookworm_amd64_hello_1.0-1_1",
                child_type=CollectionItem.Types.BARE,
                category=BareDataCategory.PACKAGE_BUILD_LOG,
                collection__isnull=True,
                artifact__isnull=True,
                data=bare_item.data,
                removed_at__isnull=True,
            ).exists()
        )

    def test_copy_item_bare_variables(self) -> None:
        """Pass additional variables while copying a bare item."""
        bare_item = self.create_build_log_placeholder()
        source_items = self.prepare_copy_item_test(
            {"child_type": LookupChildType.BARE}
        )
        self.assertEqual(len(source_items), 1)

        self.task.copy_item(
            source_items[0],
            self.target_build_logs_input,
            user=self.scenario.user,
            variables={"additional": "foo"},
        )

        self.assertTrue(
            self.target_build_logs.child_items.filter(
                name="debian_bookworm_amd64_hello_1.0-1_1",
                child_type=CollectionItem.Types.BARE,
                category=BareDataCategory.PACKAGE_BUILD_LOG,
                collection__isnull=True,
                artifact__isnull=True,
                data={**bare_item.data, "additional": "foo"},
                removed_at__isnull=True,
            ).exists()
        )

    def test_copy_item_artifact(self) -> None:
        sbuild_work_request = self.playground.create_worker_task(
            workspace=self.source_workspace
        )
        self.playground.create_group_role(
            self.scenario.workspace,
            Workspace.Roles.CONTRIBUTOR,
            users=[self.scenario.user],
        )
        artifact_item = self.create_build_log(work_request=sbuild_work_request)
        artifact = artifact_item.artifact
        assert artifact is not None
        build_log_file = artifact.files.get()
        source_items = self.prepare_copy_item_test(
            {"child_type": LookupChildType.ARTIFACT}
        )
        self.assertEqual(len(source_items), 1)
        source_file_store = self.source_scope.download_file_stores(
            build_log_file
        ).get()

        self.task.work_request.set_current()
        self.task.copy_item(
            source_items[0],
            self.target_build_logs_input,
            user=self.scenario.user,
        )

        copied_collection_item = self.target_build_logs.child_items.get(
            name=f"debian_bookworm_amd64_hello_1.0-1_{sbuild_work_request.id}",
            child_type=CollectionItem.Types.ARTIFACT,
            category=ArtifactCategory.PACKAGE_BUILD_LOG,
            collection__isnull=True,
            data={
                **artifact_item.data,
                "srcpkg_name": "hello",
                "srcpkg_version": "1.0-1",
            },
            removed_at__isnull=True,
        )
        # The artifact was copied.
        copied_artifact = copied_collection_item.artifact
        assert copied_artifact is not None
        self.assertNotEqual(copied_artifact, artifact)
        self.assertEqual(copied_artifact.category, artifact.category)
        self.assertEqual(copied_artifact.workspace, self.scenario.workspace)
        self.assertEqual(copied_artifact.data, artifact.data)
        self.assertEqual(
            copied_artifact.expiration_delay,
            artifact.expiration_delay,
        )
        self.assertNotEqual(copied_artifact.created_at, artifact.created_at)
        self.assertEqual(copied_artifact.created_by, artifact.created_by)
        self.assertEqual(
            copied_artifact.created_by_work_request,
            artifact.created_by_work_request,
        )
        self.assertEqual(copied_artifact.original_artifact, artifact)
        self.assertEqual(copied_artifact.files.get(), build_log_file)
        # The file in the copied artifact is in the same store.
        target_file_store = self.scenario.scope.upload_file_stores(
            build_log_file
        ).get()
        self.assertEqual(source_file_store, target_file_store)

    def test_copy_item_artifact_needs_unembargo(self) -> None:
        """Copying an artifact from private to public needs unembargo=True."""
        sbuild_work_request = self.playground.create_worker_task(
            workspace=self.source_workspace
        )
        self.create_build_log(work_request=sbuild_work_request)
        source_items = self.prepare_copy_item_test(
            {"child_type": LookupChildType.ARTIFACT}
        )
        self.assertEqual(len(source_items), 1)
        self.source_workspace.public = False
        self.source_workspace.save()

        with self.assertRaisesRegex(
            CannotCopy,
            f"Copying from {self.source_workspace}"
            f" to {self.scenario.workspace} "
            f"requires unembargo=True",
        ):
            self.task.copy_item(
                source_items[0],
                self.target_build_logs_input,
                user=self.scenario.user,
            )

    def test_copy_item_artifact_with_unembargo(self) -> None:
        """Copying an artifact from private to public can unembargo it."""
        sbuild_work_request = self.playground.create_worker_task(
            workspace=self.source_workspace, assign_contributor_role=True
        )
        self.playground.create_group_role(
            self.scenario.workspace,
            Workspace.Roles.CONTRIBUTOR,
            users=[sbuild_work_request.created_by],
        )
        artifact_item = self.create_build_log(work_request=sbuild_work_request)
        artifact = artifact_item.artifact
        assert artifact is not None
        source_items = self.prepare_copy_item_test(
            {"child_type": LookupChildType.ARTIFACT}
        )
        self.assertEqual(len(source_items), 1)
        self.source_workspace.public = False
        self.source_workspace.save()

        self.task.work_request.set_current()
        self.task.copy_item(
            source_items[0],
            self.target_build_logs_input,
            user=self.scenario.user,
            unembargo=True,
        )

        copied_collection_item = self.target_build_logs.child_items.get(
            name=f"debian_bookworm_amd64_hello_1.0-1_{sbuild_work_request.id}",
            child_type=CollectionItem.Types.ARTIFACT,
            category=ArtifactCategory.PACKAGE_BUILD_LOG,
            collection__isnull=True,
            data={
                **artifact_item.data,
                "srcpkg_name": "hello",
                "srcpkg_version": "1.0-1",
            },
            removed_at__isnull=True,
        )
        # The artifact was copied.
        copied_artifact = copied_collection_item.artifact
        assert copied_artifact is not None
        self.assertEqual(copied_artifact.workspace, self.scenario.workspace)

    def test_copy_item_artifact_variables(self) -> None:
        """Pass additional variables while copying an artifact."""
        sbuild_work_request = self.playground.create_worker_task(
            workspace=self.source_workspace
        )
        self.playground.create_group_role(
            self.scenario.workspace,
            Workspace.Roles.CONTRIBUTOR,
            users=[sbuild_work_request.created_by],
        )
        self.create_build_log(extra_data={"nested": {"foo": "bar"}})
        source_items = self.prepare_copy_item_test(
            {"child_type": LookupChildType.ARTIFACT}
        )
        self.assertEqual(len(source_items), 1)
        target_collection = self.playground.create_collection(
            name="test",
            category=CollectionCategory.WORKFLOW_INTERNAL,
            workspace=self.scenario.workspace,
        )

        self.task.work_request.set_current()
        self.task.copy_item(
            source_items[0],
            DBInputCollectionSingle.from_collection(
                target_collection,
                lookup="internal@collections",
                label="copies[0].target_collection",
            ),
            user=self.scenario.user,
            name_template="{vendor}_{nested_foo}",
            variables={"$nested_foo": "nested.foo"},
        )

        self.assertTrue(
            target_collection.child_items.filter(
                name="debian_bar",
                child_type=CollectionItem.Types.ARTIFACT,
                category=ArtifactCategory.PACKAGE_BUILD_LOG,
                collection__isnull=True,
                data={},
                removed_at__isnull=True,
            ).exists()
        )

    def test_copy_item_artifact_failed_to_expand_variables(self) -> None:
        """Fail if the given variables cannot be expanded."""
        self.create_build_log()
        source_items = self.prepare_copy_item_test(
            {"child_type": LookupChildType.ARTIFACT}
        )
        self.assertEqual(len(source_items), 1)

        with self.assertRaisesRegex(
            CannotCopy, r"Cannot expand variables: \{'\$key': 'nonexistent'\}"
        ):
            self.task.copy_item(
                source_items[0],
                self.target_build_logs_input,
                user=self.scenario.user,
                variables={"$key": "nonexistent"},
            )

    def test_copy_item_artifact_incomplete_file(self) -> None:
        """We cannot copy artifacts with incomplete files."""
        sbuild_work_request = self.playground.create_worker_task(
            workspace=self.source_workspace
        )
        self.playground.create_group_role(
            self.scenario.workspace,
            Workspace.Roles.CONTRIBUTOR,
            users=[sbuild_work_request.created_by],
        )
        artifact_item = self.create_build_log(skip_add_files_in_store=True)
        source_items = self.prepare_copy_item_test(
            {"child_type": LookupChildType.ARTIFACT}
        )
        self.assertEqual(len(source_items), 1)

        self.task.work_request.set_current()
        with self.assertRaisesRegex(
            CannotCopy,
            f"Cannot copy incomplete 'hello_1.0-1_amd64.buildlog' from "
            f"artifact ID {artifact_item.artifact_id}",
        ):
            self.task.copy_item(
                source_items[0],
                self.target_build_logs_input,
                user=self.scenario.user,
            )

    def test_copy_item_artifact_not_in_any_file_store(self) -> None:
        """We cannot copy artifacts that are not in any file store."""
        sbuild_work_request = self.playground.create_worker_task(
            workspace=self.source_workspace
        )
        self.playground.create_group_role(
            self.scenario.workspace,
            Workspace.Roles.CONTRIBUTOR,
            users=[sbuild_work_request.created_by],
        )
        artifact_item = self.create_build_log()
        source_items = self.prepare_copy_item_test(
            {"child_type": LookupChildType.ARTIFACT}
        )
        self.assertEqual(len(source_items), 1)
        self.source_scope.file_stores.get().files.clear()

        self.task.work_request.set_current()
        with self.assertRaisesRegex(
            CannotCopy,
            f"Cannot copy 'hello_1.0-1_amd64.buildlog' from artifact ID "
            f"{artifact_item.artifact_id}: not in any available file store",
        ):
            self.task.copy_item(
                source_items[0],
                self.target_build_logs_input,
                user=self.scenario.user,
            )

    def test_copy_item_artifact_different_file_store(self) -> None:
        """
        Copy artifact file contents from a different file store.

        We copy between scopes in order to exercise the case where the
        download and upload file stores are different.
        """
        file_store = FileStore.objects.create(
            name="test",
            backend=FileStore.BackendChoices.LOCAL,
            configuration={"base_directory": settings.DEBUSINE_STORE_DIRECTORY},
        )
        self.source_scope.file_stores.set([file_store])
        sbuild_work_request = self.playground.create_worker_task(
            workspace=self.source_workspace
        )
        self.playground.create_group_role(
            self.scenario.workspace,
            Workspace.Roles.CONTRIBUTOR,
            users=[sbuild_work_request.created_by],
        )
        build_log_contents = b"a build log\n"
        artifact_item = self.create_build_log(
            work_request=sbuild_work_request,
            build_log_contents=build_log_contents,
        )
        artifact = artifact_item.artifact
        assert artifact is not None
        build_log_file = artifact.files.get()
        source_items = self.prepare_copy_item_test(
            {"child_type": LookupChildType.ARTIFACT}
        )
        self.assertEqual(len(source_items), 1)
        source_file_store = self.source_scope.download_file_stores(
            build_log_file
        ).get()

        self.task.work_request.set_current()
        self.task.copy_item(
            source_items[0],
            self.target_build_logs_input,
            user=self.scenario.user,
        )

        copied_collection_item = self.target_build_logs.child_items.get(
            name=f"debian_bookworm_amd64_hello_1.0-1_{sbuild_work_request.id}",
            child_type=CollectionItem.Types.ARTIFACT,
            category=ArtifactCategory.PACKAGE_BUILD_LOG,
            collection__isnull=True,
            data={
                **artifact_item.data,
                "srcpkg_name": "hello",
                "srcpkg_version": "1.0-1",
            },
            removed_at__isnull=True,
        )
        copied_artifact = copied_collection_item.artifact
        assert copied_artifact is not None
        self.assertNotEqual(copied_artifact, artifact)
        self.assertQuerySetEqual(copied_artifact.files.all(), [build_log_file])
        # The file in the copied artifact is in a different store.
        target_file_store = self.scenario.scope.upload_file_stores(
            build_log_file
        ).get()
        self.assertNotEqual(source_file_store, target_file_store)
        # The file contents were copied.
        with (
            source_file_store.get_backend_object().get_stream(
                build_log_file
            ) as source_file,
            target_file_store.get_backend_object().get_stream(
                build_log_file
            ) as target_file,
        ):
            self.assertEqual(target_file.read(), source_file.read())

    def test_execute(self) -> None:
        """The task copies all the given items."""
        # Partially simulate a package build in a workflow, and add the
        # resulting artifacts to appropriate source collections as starting
        # points for copies.
        template = self.playground.create_workflow_template(
            name="test", task_name="noop"
        )
        workflow = self.playground.create_workflow(template, task_data={})
        assert workflow.internal_collection is not None
        source_package = self.playground.create_source_artifact(
            workspace=self.source_workspace
        )
        sbuild = self.playground.simulate_package_build(
            source_package, workflow=workflow
        )
        binary_package = sbuild.artifact_set.get(
            category=ArtifactCategory.BINARY_PACKAGE
        )
        workflow.internal_collection.manager.add_artifact(
            binary_package,
            user=sbuild.created_by,
            workflow=workflow,
            name="binary-amd64",
        )
        build_log = sbuild.artifact_set.get(
            category=ArtifactCategory.PACKAGE_BUILD_LOG
        )
        workflow.internal_collection.manager.add_artifact(
            build_log,
            user=sbuild.created_by,
            workflow=workflow,
            name="build-log",
            variables={
                "work_request_id": sbuild.id,
                "vendor": "debian",
                "codename": "bookworm",
                "architecture": "amd64",
            },
        )

        # Make a target suite and a task to perform the copies.
        target_suite = self.playground.create_collection(
            name="trixie",
            category=CollectionCategory.SUITE,
            workspace=self.scenario.workspace,
        )
        binary_package_lookup = LookupMultiple.model_validate(
            {
                "collection": "internal@collections",
                "category": ArtifactCategory.BINARY_PACKAGE,
            }
        )
        copy_collection_items = workflow.create_child_server(
            task_name="copycollectionitems",
            task_data=CopyCollectionItemsData(
                copies=[
                    CopyCollectionItemsCopies(
                        source_items=binary_package_lookup,
                        target_collection=target_suite.id,
                        variables={
                            "component": "main",
                            "section": "python",
                            "priority": "optional",
                        },
                    ),
                    CopyCollectionItemsCopies(
                        source_items=LookupMultiple.model_validate(
                            {
                                "collection": "internal@collections",
                                "name": "build-log",
                            }
                        ),
                        target_collection=self.target_build_logs.id,
                    ),
                ]
            ),
        )
        self.playground.create_group_role(
            self.scenario.workspace,
            Workspace.Roles.CONTRIBUTOR,
            users=[copy_collection_items.created_by],
        )
        self.playground.advance_work_request(
            copy_collection_items, assign_new_worker=True
        )
        task = copy_collection_items.get_task()

        with self.playground.assign_role(
            self.source_workspace,
            copy_collection_items.created_by,
            Workspace.Roles.CONTRIBUTOR,
        ):
            self.assertEqual(task.execute(), WorkRequest.Results.SUCCESS)

        debug_logs_artifact = copy_collection_items.artifact_set.get(
            category=ArtifactCategory.WORK_REQUEST_DEBUG_LOGS
        )
        copy_log = debug_logs_artifact.fileinartifact_set.get()
        self.assertEqual(copy_log.path, "copy-collection-items.log")
        fileobj = copy_log.file
        file_backend = self.source_workspace.scope.download_file_backend(
            fileobj
        )
        with file_backend.get_stream(fileobj) as file:
            self.assertEqual(
                file.read().decode(),
                dedent(
                    f"""\
                    Copying to {target_suite}: {binary_package}
                    Copying to {self.target_build_logs}: {build_log}
                    """
                ),
            )

        # The target collections contain the requested items.
        self.assertTrue(
            target_suite.child_items.filter(
                name="hello_1.0-1_all",
                child_type=CollectionItem.Types.ARTIFACT,
                category=ArtifactCategory.BINARY_PACKAGE,
                artifact__isnull=False,
                data={
                    "srcpkg_name": "hello",
                    "srcpkg_version": "1.0-1",
                    "package": "hello",
                    "version": "1.0-1",
                    "architecture": "all",
                    "component": "main",
                    "section": "python",
                    "priority": "optional",
                },
                removed_at__isnull=True,
            ).exists()
        )
        self.assertTrue(
            self.target_build_logs.child_items.filter(
                name=f"debian_bookworm_amd64_hello_1.0-1_{sbuild.id}",
                child_type=CollectionItem.Types.ARTIFACT,
                category=ArtifactCategory.PACKAGE_BUILD_LOG,
                artifact__isnull=False,
                data__work_request_id=sbuild.id,
                removed_at__isnull=True,
            ).exists()
        )

    def test_execute_copy_fails(self) -> None:
        self.create_build_log_placeholder()
        source_items = LookupMultiple.model_validate(
            {
                "collection": self.source_build_logs.id,
                "child_type": LookupChildType.BARE,
            }
        )
        self.task_data = {
            "copies": [
                CopyCollectionItemsCopies(
                    source_items=source_items,
                    target_collection=self.target_build_logs.id,
                )
            ]
        }
        copy_collection_items = self.work_request

        self.source_workspace.public = False
        self.source_workspace.save()
        self.playground.create_group_role(
            self.scenario.workspace,
            Workspace.Roles.CONTRIBUTOR,
            users=[copy_collection_items.created_by],
        )
        self.playground.advance_work_request(
            copy_collection_items, assign_new_worker=True
        )

        with (
            self.playground.assign_role(
                self.source_workspace,
                copy_collection_items.created_by,
                Workspace.Roles.CONTRIBUTOR,
            ),
            self.assertRaises(CannotCopy),
        ):
            self.task.execute()

        assert self.task._debug_log_files_directory is not None
        self.assertEqual(
            Path(
                self.task._debug_log_files_directory.name,
                "copy-collection-items.log",
            ).read_text(),
            f"Copy failed: Copying from {self.source_workspace} to "
            f"{self.scenario.workspace} requires unembargo=True\n",
        )

    def test_get_label(self) -> None:
        """Test get_label."""
        self.playground.create_collection(
            name="bookworm", category=CollectionCategory.SUITE
        )
        self.playground.create_collection(
            name="trixie", category=CollectionCategory.SUITE
        )
        self.task_data = {
            "copies": [
                {
                    "source_items": {"collection": 1},
                    "target_collection": "bookworm@debian:suite",
                },
                {
                    "source_items": {"collection": 1},
                    "target_collection": "trixie@debian:suite",
                },
            ]
        }

        self.assertEqual(
            self.task.get_label(),
            "copy to bookworm@debian:suite, trixie@debian:suite",
        )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for scheduler."""

import datetime as dt
import logging
from contextlib import ExitStack
from typing import Any, TYPE_CHECKING, override
from unittest import mock

from celery import states
from celery.contrib.testing.app import TestApp, setup_default_app
from channels.db import database_sync_to_async
from django.db import OperationalError, transaction
from django.db.models import QuerySet
from django.test import override_settings
from django.utils import timezone
from django_celery_results.models import TaskResult

import debusine.tasks.tags as ttags
import debusine.worker.tags as wtags
from debusine.artifacts.models import CollectionCategory
from debusine.assets import KeyPurpose
from debusine.db.context import context
from debusine.db.models import CollectionItem, WorkRequest, Worker
from debusine.db.models.lookups import LookupCache
from debusine.db.models.work_requests import workflow_ancestors
from debusine.db.playground import scenarios
from debusine.db.tasks.mock_resolver import MockFieldResolver
from debusine.db.tests.utils import RunInParallelTransaction
from debusine.project.celery import (
    TaskWithBetterArgumentRepresentation,
    celery_tasks_in_progress,
    get_apply_async_tasks,
)
from debusine.server.scheduler import (
    get_provided_worker_tags,
    schedule,
    schedule_for_worker,
    schedule_soon,
    schedule_task,
)
from debusine.server.tasks import ServerNoop
from debusine.server.tasks.tests.helpers import SampleBaseWaitTask
from debusine.server.tasks.wait.models import ConfirmData, DelayData
from debusine.server.workflows.models import WorkRequestWorkflowData
from debusine.signing.tasks.models import GenerateKeyData
from debusine.tasks import Lintian, Sbuild
from debusine.tasks.models import (
    ActionSkipIfLookupResultChanged,
    BackendType,
    BaseDynamicTaskData,
    BaseTaskData,
    Confirmation,
    EventReactions,
    LintianData,
    LintianInput,
    OutputData,
    OutputDataError,
    SbuildBuildComponent,
    SbuildData,
    SbuildInput,
    WorkerType,
)
from debusine.test.django import (
    PlaygroundTestCase,
    TestCase,
    TransactionTestCase,
    preserve_db_task_registry,
)
from debusine.worker.system_information import native_architecture

if TYPE_CHECKING:
    from debusine.test.playground import Playground


class SchedulerTestCase(PlaygroundTestCase):
    """Additional methods to help test scheduler related functions."""

    scenario = scenarios.DefaultSbuild()

    SAMPLE_TASK_DATA = SbuildData(
        input=SbuildInput(source_artifact=10),
        environment="debian/match:codename=sid",
        build_architecture="amd64",
        build_components=[
            SbuildBuildComponent("any"),
            SbuildBuildComponent.ANY,
            SbuildBuildComponent.ALL,
        ],
    )

    playground: "Playground"

    @override
    def setUp(self) -> None:
        super().setUp()
        self.task_data = self.SAMPLE_TASK_DATA.model_copy()
        self.task_data.input.source_artifact = self.scenario.source.id

        self.resolver = MockFieldResolver(
            extra_repositories=None,
            debusine_fqdn="debusine.example.net",
            task_configuration=None,
        )
        self.resolver.set_artifact("environment", self.scenario.environment_sid)
        self.resolver.set_artifact(
            "input_source_artifact", self.scenario.source
        )
        self.resolver.set_field_artifacts("input_binary_artifacts")
        self.resolver.set_field_artifacts("input_extra_binary_artifacts")
        self.resolver.set_collection(
            "task_configuration",
            self.scenario.default_task_configuration_collection,
        )

    def patch_field_resolver(self) -> None:
        """Patch :py:class:`debusine.server.scheduler.TaskDatabase`."""
        field_resolver_patcher = mock.patch(
            "debusine.db.tasks.base.DBFieldResolver",
            side_effect=lambda _: self.resolver,
        )
        field_resolver_patcher.start()
        self.addCleanup(field_resolver_patcher.stop)

    @transaction.atomic
    def _create_worker(
        self, *, enabled: bool, worker_type: WorkerType = WorkerType.EXTERNAL
    ) -> Worker:
        """Create and return a Worker."""
        token = self.playground.create_worker_token(
            enabled=enabled, set_worker=False
        )
        worker = Worker.objects.create_with_fqdn(
            'worker.lan', token=token, worker_type=worker_type
        )

        worker.mark_connected()

        worker.set_dynamic_metadata(
            {
                "system:worker_type": worker_type,
                "system:native_architecture": "amd64",
                "system:architectures": ["amd64"],
                "sbuild:version": Sbuild.TASK_VERSION,
                "sbuild:available": True,
                "signing:generatekey:version": 1,
                "executor:unshare:available": True,
            }
        )

        return worker

    def create_sample_sbuild_work_request(
        self,
        require_scheduler_tags: bool = True,
        assign_to: Worker | None = None,
        parent: WorkRequest | None = None,
    ) -> WorkRequest:
        """Create and return a WorkRequest."""
        work_request = self.playground.create_worker_task(
            task_name="sbuild", task_data=self.SAMPLE_TASK_DATA, parent=parent
        )
        with ExitStack() as stack:
            if not require_scheduler_tags:
                stack.enter_context(
                    mock.patch(
                        "debusine.db.tasks.base.DBProxyTask"
                        ".compute_scheduler_tags",
                        return_value=(
                            ttags.TaskProvidedTags(),
                            wtags.TaskRequiredTags(),
                        ),
                    )
                )
            self.playground.advance_work_request(
                work_request, mark_pending=True, worker=assign_to
            )
        return work_request

    def set_tasks_allowlist(self, allowlist: list[str]) -> None:
        """Set tasks_allowlist in worker's static metadata."""
        assert hasattr(self, "worker")
        self.worker.static_metadata["tasks_allowlist"] = allowlist
        self.worker.save()

    def set_tasks_denylist(self, denylist: list[str]) -> None:
        """Set tasks_denylist in worker's static metadata."""
        assert hasattr(self, "worker")
        self.worker.static_metadata["tasks_denylist"] = denylist
        self.worker.save()

    def create_wait_delay_task(
        self,
        delay_offset: dt.timedelta,
    ) -> WorkRequest:
        """Create a Wait/Delay task in a workflow."""
        parent = self.playground.create_workflow(mark_running=True)
        return parent.create_child_wait(
            task_name="delay",
            task_data=DelayData(delay_until=timezone.now() + delay_offset),
            workflow_data=WorkRequestWorkflowData(needs_input=False),
        )


class SchedulerTests(SchedulerTestCase, TransactionTestCase):
    """Test schedule() function."""

    @override
    def setUp(self) -> None:
        super().setUp()
        self.work_request_1 = self.create_sample_sbuild_work_request()
        self.work_request_2 = self.create_sample_sbuild_work_request()
        self.patch_field_resolver()

    def test_no_work_request_available(self) -> None:
        """schedule() returns nothing when no work request is available."""
        self._create_worker(enabled=True)
        WorkRequest.objects.all().delete()

        self.assertEqual(schedule(), [])

    def test_only_unhandled_work_requests_available(self) -> None:
        """schedule() returns nothing when no work request is available."""
        self._create_worker(enabled=True)
        WorkRequest.objects.all().delete()

        self.playground.create_internal_task(
            mark_pending=True, task_name="unhandled", validate=False
        )

        self.assertEqual(schedule(), [])

    def test_synchronization_points(self) -> None:
        """schedule() marks pending synchronization points as completed."""
        WorkRequest.objects.all().delete()

        parent = self.playground.create_workflow(mark_running=True)

        pending = [
            WorkRequest.objects.create_synchronization_point(
                parent=parent, step=f"pending{i}"
            )
            for i in range(2)
        ]
        for wr in pending:
            wr.mark_pending()
        self.playground.advance_work_request(
            WorkRequest.objects.create_synchronization_point(
                parent=parent,
                step="completed",
            ),
            result=WorkRequest.Results.SUCCESS,
        )

        self.assertCountEqual(schedule(), pending)
        for wr in pending:
            wr.refresh_from_db()
            self.assertEqual(wr.status, WorkRequest.Statuses.COMPLETED)
            self.assertEqual(wr.result, WorkRequest.Results.SUCCESS)

    @override_settings(DISABLE_AUTOMATIC_SCHEDULING=True)
    def test_worker_connected_without_a_work_request(self) -> None:
        """schedule() goes over all workers."""
        self.work_request_1.scheduler_tags_required = []
        self.work_request_1.save()
        self.work_request_2.scheduler_tags_required = []
        self.work_request_2.save()
        self._create_worker(enabled=True)
        self._create_worker(enabled=True)

        result = schedule()

        self.assertCountEqual(
            result, {self.work_request_1, self.work_request_2}
        )

    def test_worker_connected_with_a_work_request(self) -> None:
        """schedule() skips a worker with assigned work requests."""
        # Create worker_1 and assigns to work_request_1
        worker_1 = self._create_worker(enabled=True)
        self.work_request_1.assign_worker(worker_1)

        # schedule() has no worker available
        self.assertEqual(schedule(), [])

    def test_worker_not_connected(self) -> None:
        """schedule() skips the workers that are not connected."""
        worker = self._create_worker(enabled=True)
        worker.connected_at = None
        worker.save()

        self.assertEqual(schedule(), [])

    def test_worker_not_enabled(self) -> None:
        """schedule() skips workers that are not enabled."""
        worker = self._create_worker(enabled=False)
        worker.mark_connected()

        self.assertEqual(schedule(), [])

    @override_settings(DISABLE_AUTOMATIC_SCHEDULING=True)
    def test_wait_delay_task_past(self) -> None:
        """schedule() marks expired Wait/Delay tasks as completed."""
        work_request = self.create_wait_delay_task(dt.timedelta(seconds=-1))
        work_request.mark_pending()

        result = schedule()

        self.assertEqual(result, [work_request])
        work_request.refresh_from_db()
        self.assertEqual(work_request.status, WorkRequest.Statuses.COMPLETED)
        self.assertEqual(work_request.result, WorkRequest.Results.SUCCESS)
        self.assertIsNone(work_request.output_data)

    @override_settings(DISABLE_AUTOMATIC_SCHEDULING=True)
    def test_wait_delay_task_future(self) -> None:
        """schedule() leaves non-expired Wait/Delay tasks running."""
        work_request = self.create_wait_delay_task(dt.timedelta(hours=1))
        work_request.mark_pending()

        result = schedule()

        self.assertEqual(result, [work_request])
        work_request.refresh_from_db()
        self.assertEqual(work_request.status, WorkRequest.Statuses.RUNNING)

    @override_settings(DISABLE_AUTOMATIC_SCHEDULING=True)
    def test_wait_delay_task_blocked(self) -> None:
        """schedule() skips blocked Wait/Delay tasks."""
        work_request = self.create_wait_delay_task(dt.timedelta(seconds=-1))

        result = schedule()

        self.assertEqual(result, [])
        work_request.refresh_from_db()
        self.assertEqual(work_request.status, WorkRequest.Statuses.BLOCKED)

    def create_wait_confirm_task(
        self, *, auto_confirm_if_no_failure: bool
    ) -> WorkRequest:
        """Create a Wait/Confirm task in a workflow."""
        parent = self.playground.create_workflow(mark_running=True)
        return parent.create_child_wait(
            task_name="confirm",
            task_data=ConfirmData(
                auto_confirm_if_no_failure=auto_confirm_if_no_failure
            ),
            workflow_data=WorkRequestWorkflowData(needs_input=True),
        )

    @override_settings(DISABLE_AUTOMATIC_SCHEDULING=True)
    def test_wait_confirm_task_auto_confirm_dep_success(self) -> None:
        """schedule() handles auto-confirming tasks where all deps succeeded."""
        work_request = self.create_wait_confirm_task(
            auto_confirm_if_no_failure=True
        )
        assert work_request.parent is not None
        dep = work_request.parent.create_child_worker(task_name="noop")
        work_request.add_dependency(dep)
        self.playground.advance_work_request(
            dep, result=WorkRequest.Results.SUCCESS
        )

        result = schedule()

        self.assertEqual(result, [work_request])
        work_request.refresh_from_db()
        self.assertEqual(work_request.status, WorkRequest.Statuses.COMPLETED)
        self.assertEqual(work_request.result, WorkRequest.Results.SUCCESS)
        self.assertEqual(
            work_request.output_data,
            OutputData(confirmation=Confirmation(confirmed=True)),
        )

    @override_settings(DISABLE_AUTOMATIC_SCHEDULING=True)
    def test_wait_confirm_task_auto_confirm_dep_failure(self) -> None:
        """schedule() skips auto-confirming tasks where a dep failed."""
        work_request = self.create_wait_confirm_task(
            auto_confirm_if_no_failure=True
        )
        assert work_request.parent is not None
        dep = work_request.parent.create_child_worker(
            task_name="noop",
            workflow_data=WorkRequestWorkflowData(allow_failure=True),
        )
        work_request.add_dependency(dep)
        self.playground.advance_work_request(
            dep, result=WorkRequest.Results.FAILURE
        )

        result = schedule()

        self.assertEqual(result, [work_request])
        work_request.refresh_from_db()
        self.assertEqual(work_request.status, WorkRequest.Statuses.RUNNING)

    @override_settings(DISABLE_AUTOMATIC_SCHEDULING=True)
    def test_wait_confirm_task_non_auto_confirm(self) -> None:
        """schedule() skips non-auto-confirming Confirm tasks."""
        work_request = self.create_wait_confirm_task(
            auto_confirm_if_no_failure=False
        )
        work_request.mark_pending()

        result = schedule()

        self.assertEqual(result, [work_request])
        work_request.refresh_from_db()
        self.assertEqual(work_request.status, WorkRequest.Statuses.RUNNING)

    @override_settings(DISABLE_AUTOMATIC_SCHEDULING=True)
    def test_wait_confirm_task_blocked(self) -> None:
        """schedule() skips blocked Wait/Confirm tasks."""
        work_request = self.create_wait_confirm_task(
            auto_confirm_if_no_failure=True
        )

        result = schedule()

        self.assertEqual(result, [])
        work_request.refresh_from_db()
        self.assertEqual(work_request.status, WorkRequest.Statuses.BLOCKED)

    @override_settings(DISABLE_AUTOMATIC_SCHEDULING=True)
    @preserve_db_task_registry()
    def test_wait_other_task(self) -> None:
        """schedule() leaves non-Delay Wait tasks running."""

        class WaitNoop(SampleBaseWaitTask[BaseTaskData, BaseDynamicTaskData]):
            TASK_VERSION = 1

            @override
            def _execute(self) -> WorkRequest.Results:
                raise NotImplementedError()

        parent = self.playground.create_workflow(mark_running=True)
        work_request = parent.create_child_wait(
            task_name="waitnoop",
            workflow_data=WorkRequestWorkflowData(needs_input=False),
        )
        work_request.mark_pending()

        result = schedule()

        self.assertEqual(result, [work_request])
        work_request.refresh_from_db()
        self.assertEqual(work_request.status, WorkRequest.Statuses.RUNNING)

    def test_schedule_internal_locks_workflows(self) -> None:
        with transaction.atomic():
            work_request = self.create_wait_delay_task(dt.timedelta(hours=1))
            self.playground.advance_work_request(
                work_request, mark_pending=True
            )
            assert work_request.parent is not None
            noop = work_request.parent.create_child_worker(task_name="noop")
            unrelated = self.playground.create_worker_task()

        with (
            transaction.atomic(),
            mock.patch("debusine.server.scheduler.update_workflows"),
        ):
            schedule()

            work_request.refresh_from_db()
            self.assertEqual(work_request.status, WorkRequest.Statuses.RUNNING)
            self.assert_work_requests_locked(
                locked=[work_request.parent, work_request, noop],
                unlocked=[unrelated],
            )

    def test_schedule_internal_workflow_already_locked(self) -> None:
        with transaction.atomic():
            work_request = self.create_wait_delay_task(dt.timedelta(hours=1))
            self.playground.advance_work_request(
                work_request, mark_pending=True
            )
            assert work_request.parent is not None
            noop = work_request.parent.create_child_worker(task_name="noop")

        thread = RunInParallelTransaction(
            lambda: WorkRequest.objects.select_for_update().get(id=noop.id)
        )
        thread.start_transaction()

        try:
            with (
                mock.patch("debusine.server.scheduler.update_workflows"),
                self.assertLogsContains(
                    "[SCHEDULER] Failed to lock workflows with pending internal"
                    " work requests",
                    logger="debusine.server.scheduler",
                    level=logging.DEBUG,
                ),
            ):
                schedule()
        finally:
            thread.stop_transaction()

    @override_settings(DISABLE_AUTOMATIC_SCHEDULING=True)
    def test_complete_wait_tasks_locks_workflows(self) -> None:
        with transaction.atomic():
            work_request = self.create_wait_delay_task(dt.timedelta(seconds=-1))
            self.playground.advance_work_request(
                work_request, mark_running=True
            )
            assert work_request.parent is not None
            noop = work_request.parent.create_child_worker(task_name="noop")
            unrelated = self.playground.create_worker_task()

        with (
            transaction.atomic(),
            mock.patch("debusine.server.scheduler.update_workflows"),
        ):
            schedule()

            work_request.refresh_from_db()
            self.assertEqual(
                work_request.status, WorkRequest.Statuses.COMPLETED
            )
            self.assertEqual(work_request.result, WorkRequest.Results.SUCCESS)
            self.assert_work_requests_locked(
                locked=[work_request.parent, work_request, noop],
                unlocked=[unrelated],
            )

    def test_complete_wait_tasks_workflow_already_locked(self) -> None:
        with transaction.atomic():
            work_request = self.create_wait_delay_task(dt.timedelta(seconds=-1))
            self.playground.advance_work_request(
                work_request, mark_running=True
            )
            assert work_request.parent is not None
            noop = work_request.parent.create_child_worker(task_name="noop")

        thread = RunInParallelTransaction(
            lambda: WorkRequest.objects.select_for_update().get(id=noop.id)
        )
        thread.start_transaction()

        try:
            with (
                mock.patch("debusine.server.scheduler.update_workflows"),
                self.assertLogsContains(
                    "[SCHEDULER] Failed to lock workflows with pending internal"
                    " work requests",
                    logger="debusine.server.scheduler",
                    level=logging.DEBUG,
                ),
            ):
                schedule()
        finally:
            thread.stop_transaction()


class ScheduleForWorkerTests(SchedulerTestCase, TestCase):
    """Test schedule_for_worker()."""

    @override
    def setUp(self) -> None:
        super().setUp()
        self.worker = self._create_worker(enabled=True)
        self.patch_field_resolver()
        self.enterContext(override_settings(FEATURE_TAG_BASED_SCHEDULING=False))

    def test_no_work_request(self) -> None:
        """schedule_for_worker() returns None when nothing to do."""
        self.assertIsNone(self.playground.schedule_for_worker(self.worker))

    def test_skips_assigned_work_requests(self) -> None:
        """schedule_for_worker() doesn't pick assigned work requests."""
        # Create a running work request
        worker_2 = self._create_worker(enabled=True)
        work_request_1 = self.create_sample_sbuild_work_request()
        work_request_1.assign_worker(worker_2)

        result = self.playground.schedule_for_worker(self.worker)

        self.assertIsNone(result)

    def test_skips_non_pending_work_requests(self) -> None:
        """schedule_for_worker() skips non-pending work requests."""
        # Create a completed work request
        worker_2 = self._create_worker(enabled=True)
        work_request_1 = self.create_sample_sbuild_work_request(
            assign_to=worker_2
        )
        work_request_1.mark_running()
        work_request_1.mark_completed(WorkRequest.Results.SUCCESS)
        # Create a running work request
        work_request_2 = self.create_sample_sbuild_work_request(
            assign_to=worker_2
        )
        work_request_2.mark_running()
        # Create an aborted work request
        work_request_2 = self.create_sample_sbuild_work_request()
        work_request_2.mark_aborted()

        result = self.playground.schedule_for_worker(self.worker)

        self.assertIsNone(result)

    @override_settings(DISABLE_AUTOMATIC_SCHEDULING=True)
    def test_skips_work_request_based_on_worker_metadata(self) -> None:
        wr = self.create_sample_sbuild_work_request()
        self.worker.set_dynamic_metadata({})
        self.worker.save()
        wr.refresh_from_db()
        self.assertIsNone(wr.worker)

        result = self.playground.schedule_for_worker(self.worker)
        self.assertIsNone(result)

    def test_skips_work_request_if_lookup_result_changed(self) -> None:
        """Skip work request if an action requests it."""
        collection = self.playground.create_collection(
            "test", CollectionCategory.TEST
        )
        lookup = f"test@{CollectionCategory.TEST}/name:foo"
        work_request = self.create_sample_sbuild_work_request(
            require_scheduler_tags=False
        )
        work_request.event_reactions = EventReactions(
            on_assignment=[
                ActionSkipIfLookupResultChanged(
                    lookup=lookup, collection_item_id=None
                )
            ]
        )
        work_request.save()
        CollectionItem.objects.create_from_artifact(
            self.playground.create_artifact()[0],
            parent_collection=collection,
            name="foo",
            data={},
            created_by_user=self.playground.get_default_user(),
        )

        with self.assertLogs(
            "debusine.server.scheduler", level=logging.DEBUG
        ) as log:
            result = self.playground.schedule_for_worker(self.worker)
        work_request.refresh_from_db()

        self.assertIn(
            f"DEBUG:debusine.server.scheduler:"
            f"Skipping work request {work_request} due to on_assignment action",
            log.output,
        )
        self.assertIsNone(result)
        self.assertEqual(work_request.status, WorkRequest.Statuses.COMPLETED)
        self.assertEqual(work_request.result, WorkRequest.Results.SKIPPED)
        self.assertEqual(
            work_request.output_data,
            OutputData(skip_reason=f"Result of lookup {lookup!r} changed"),
        )

    def test_with_pending_unassigned_work_request(self) -> None:
        """schedule_for_worker() picks up the pending work request."""
        with override_settings(DISABLE_AUTOMATIC_SCHEDULING=True):
            work_request = self.create_sample_sbuild_work_request(
                require_scheduler_tags=False
            )

        with self.captureOnCommitCallbacks() as on_commit:
            result = self.playground.schedule_for_worker(self.worker)

        assert isinstance(result, WorkRequest)
        self.assertEqual(result, work_request)
        self.assertEqual(result.worker, self.worker)

        # Assigning the work request does not cause another invocation of the
        # scheduler.
        self.assertNotIn(schedule_task, get_apply_async_tasks(on_commit))

        # Lookups from computing dynamic data are not cached.
        self.assertEqual(context.lookup_cache, LookupCache())

    def test_with_failing_configure_call(self) -> None:
        """schedule_for_worker() skips invalid data and marks it as an error."""
        work_request = self.create_sample_sbuild_work_request(
            require_scheduler_tags=False
        )
        # invalid: it's missing required properties
        work_request.configured_task_data = {}
        work_request.save()

        with self.assertLogs(
            "debusine.server.scheduler", level=logging.WARNING
        ) as log:
            result = schedule_for_worker(self.worker, set())
        work_request.refresh_from_db()

        expected_message = (
            "Failed to configure: 3 validation errors for SbuildData"
        )
        self.assertIn(
            f"ERROR:debusine.server.scheduler:"
            f"Error assigning work request "
            f"{work_request.task_type}/{work_request.task_name} "
            f"({work_request.id}): {expected_message}",
            "\n".join(log.output),
        )
        self.assertIsNone(result)
        self.assertEqual(work_request.status, WorkRequest.Statuses.COMPLETED)
        self.assertEqual(work_request.result, WorkRequest.Results.ERROR)
        assert work_request.output_data is not None
        self.assertIsInstance(work_request.output_data, OutputData)
        assert work_request.output_data.errors is not None
        self.assertEqual(len(work_request.output_data.errors), 1)
        self.assertTrue(
            work_request.output_data.errors[0].message.startswith(
                expected_message
            )
        )
        self.assertEqual(
            work_request.output_data.errors[0].code, "configure-failed"
        )

    def test_with_failing_compute_dynamic_data(self) -> None:
        """schedule_for_worker() handles failure to compute dynamic data."""
        work_request = self.create_sample_sbuild_work_request(
            require_scheduler_tags=False
        )
        work_request.task_data["backend"] = BackendType.UNSHARE
        work_request.task_data["environment"] = "bad-lookup"
        work_request.save()
        self.worker.dynamic_metadata["executor:unshare:available"] = True

        with (
            mock.patch(
                "debusine.server.scheduler.WorkRequest.assign_worker",
                side_effect=LookupError("expected failure"),
            ),
            self.assertLogs(
                "debusine.server.scheduler", level=logging.WARNING
            ) as log,
        ):
            result = schedule_for_worker(self.worker, set())
        work_request.refresh_from_db()

        expected_message = "Failed to compute dynamic data: expected failure"
        self.assertEqual(
            log.output,
            [
                f"ERROR:debusine.server.scheduler:"
                f"Error assigning work request "
                f"{work_request.task_type}/{work_request.task_name} "
                f"({work_request.id}): {expected_message}"
            ],
        )

        self.assertIsNone(result)
        self.assertEqual(work_request.status, WorkRequest.Statuses.COMPLETED)
        self.assertEqual(work_request.result, WorkRequest.Results.ERROR)
        self.assertEqual(
            work_request.output_data,
            OutputData(
                errors=[
                    OutputDataError(
                        message=expected_message, code="dynamic-data-failed"
                    )
                ]
            ),
        )

    @override_settings(DISABLE_AUTOMATIC_SCHEDULING=True)
    def test_computes_dynamic_data(
        self,
    ) -> None:
        """schedule_for_worker() uses worker arch to compute dynamic data."""
        self.patch_field_resolver()
        self.worker.dynamic_metadata.update(
            {
                "system:native_architecture": "i386",
                "system:architectures": ["amd64", "i386"],
                "lintian:version": Lintian.TASK_VERSION,
                "provided_tags": [
                    "worker:executor:unshare",
                    "worker:task:worker:lintian:version:1",
                    "worker:type:external",
                    wtags.BUILD_ARCH_PREFIX + "amd64",
                    wtags.BUILD_ARCH_PREFIX + "i386",
                ],
            }
        )
        # Lintian tasks don't care what architecture they run on, so they
        # defer to the worker's native architecture.
        work_request = self.playground.create_worker_task(
            task_name="lintian",
            task_data=LintianData(
                input=LintianInput(source_artifact=1),
                environment="debian/match:codename=bookworm",
            ),
            mark_pending=True,
        )
        work_request.save()

        result = self.playground.schedule_for_worker(self.worker)
        assert isinstance(result, WorkRequest)
        self.assertEqual(result, work_request)
        self.assertEqual(result.worker, self.worker)

        task_config_collection = (
            self.scenario.default_task_configuration_collection
        )

        self.assertEqual(
            result.configured_task_data,
            work_request.task_data
            | {"task_configuration": task_config_collection.pk},
        )
        assert result.dynamic_task_data is not None
        dynamic_data = result.dynamic_task_data.copy()
        task_input_fields = dynamic_data.pop("task_input_fields")
        self.assertEqual(
            dynamic_data,
            {
                "input_source_artifact_id": self.scenario.source.id,
                "input_binary_artifacts_ids": [],
                "environment_id": self.scenario.environment_sid.id,
                "subject": "hello",
                "parameter_summary": "hello_1.0-1",
                "runtime_context": "binary-all+binary-any+source:sid",
                "configuration_context": "sid",
            },
        )
        self.assertCountEqual(
            task_input_fields.keys(),
            [
                "build_architecture",
                "environment",
                "input_binary_artifacts",
                "input_source_artifact",
                "task_configuration",
            ],
        )

    @override_settings(DISABLE_AUTOMATIC_SCHEDULING=True)
    def test_skips_tasks_in_deny_list(self) -> None:
        """schedule_for_worker() doesn't schedule task in denylist."""
        self.create_sample_sbuild_work_request()  # is sbuild task
        self.set_tasks_denylist(["foobar", "sbuild", "baz"])

        self.assertIsNone(self.playground.schedule_for_worker(self.worker))

    def test_skips_tasks_not_in_allow_list(self) -> None:
        """schedule_for_worker() doesn't schedule task in denylist."""
        self.create_sample_sbuild_work_request()  # is sbuild task
        self.set_tasks_allowlist(["foobar", "baz"])

        self.assertIsNone(self.playground.schedule_for_worker(self.worker))

    @override_settings(DISABLE_AUTOMATIC_SCHEDULING=True)
    def test_allowlist_takes_precedence_over_denylist(self) -> None:
        """A task that is in both allow/denylist is allowed."""
        work_request = self.create_sample_sbuild_work_request(
            require_scheduler_tags=False
        )
        self.set_tasks_allowlist(["sbuild"])
        self.set_tasks_denylist(["sbuild"])

        result = self.playground.schedule_for_worker(self.worker)

        self.assertEqual(work_request, result)

    @override_settings(DISABLE_AUTOMATIC_SCHEDULING=True)
    def test_highest_priority_work_request_is_picked_first(self) -> None:
        """schedule_for_worker() picks the highest-priority work request."""
        list_of_work_requests = []
        for priority_base, priority_adjustment in (
            (0, 0),
            (0, 10),
            (10, 0),
            (10, 1),
        ):
            work_request = self.create_sample_sbuild_work_request(
                require_scheduler_tags=False
            )
            work_request.priority_base = priority_base
            work_request.priority_adjustment = priority_adjustment
            work_request.save()
            list_of_work_requests.append(work_request)

        result = self.playground.schedule_for_worker(self.worker)

        self.assertEqual(list_of_work_requests[3], result)

    @override_settings(DISABLE_AUTOMATIC_SCHEDULING=True)
    def test_oldest_work_request_is_picked_first(self) -> None:
        """schedule_for_worker() picks the oldest work request."""
        now = timezone.now()
        list_of_work_requests = []
        for offset in (1, 0, 2):  # The second object created will be the oldest
            work_request = self.create_sample_sbuild_work_request(
                require_scheduler_tags=False
            )
            work_request.created_at = now + dt.timedelta(seconds=offset)
            work_request.save()
            list_of_work_requests.append(work_request)

        result = self.playground.schedule_for_worker(self.worker)

        self.assertEqual(list_of_work_requests[1], result)

    def test_signing_work_request(self) -> None:
        """schedule_for_worker() assigns a signing work request."""
        signing_worker = self._create_worker(
            enabled=True,
            worker_type=WorkerType.SIGNING,
        )
        signing_worker.static_metadata = {
            "provided_tags": [
                wtags.WORKER_TYPE_SIGNING,
                wtags.TASK_PREFIX + "signing:generatekey:version:1",
                wtags.BUILD_ARCH_PREFIX + "amd64",
            ]
        }
        signing_worker.save()
        with override_settings(DISABLE_AUTOMATIC_SCHEDULING=True):
            work_request = self.playground.create_signing_task(
                task_name="generatekey",
                task_data=GenerateKeyData(
                    purpose=KeyPurpose.UEFI, description="A UEFI key"
                ),
                mark_pending=True,
            )

        result = self.playground.schedule_for_worker(signing_worker)

        assert isinstance(result, WorkRequest)
        self.assertEqual(result, work_request)
        self.assertEqual(result.worker, signing_worker)

    def assert_schedule_for_worker_return_none(
        self, status: WorkRequest.Statuses
    ) -> None:
        """Schedule_for_worker returns None if it has WorkRequest in status."""
        work_request = self.create_sample_sbuild_work_request(
            assign_to=self.worker
        )
        self.create_sample_sbuild_work_request()

        work_request.status = status
        work_request.save()

        self.assertIsNone(self.playground.schedule_for_worker(self.worker))

    def test_schedule_for_worker_already_pending(self) -> None:
        """Do not assign a WorkRequest if the Worker has enough pending."""
        self.assert_schedule_for_worker_return_none(
            WorkRequest.Statuses.PENDING
        )

    def test_schedule_for_worker_already_running(self) -> None:
        """Do not assign a WorkRequest if the Worker has enough running."""
        self.assert_schedule_for_worker_return_none(
            WorkRequest.Statuses.RUNNING
        )

    @mock.patch("debusine.server.scheduler.run_server_task.apply_async")
    def test_schedule_for_worker_concurrency(
        self,
        mock_apply_async: mock.MagicMock,  # noqa: ARG002, U100
    ) -> None:
        """Consider a Worker's concurrency when assigning a WorkRequest."""
        worker = Worker.objects.get_or_create_celery()
        worker.concurrency = 3
        worker.dynamic_metadata = {
            "system:architectures": [native_architecture()],
            "system:worker_type": WorkerType.CELERY,
            "server:servernoop:version": ServerNoop.TASK_VERSION,
            "provided_tags": [
                wtags.WORKER_TYPE_CELERY,
                wtags.TASK_PREFIX + "server:servernoop:version:1",
            ],
        }

        for _ in range(3):
            work_request = self.playground.create_server_task(
                task_name="servernoop", mark_pending=True
            )
            self.assertEqual(
                work_request, self.playground.schedule_for_worker(worker)
            )

        work_request = self.playground.create_server_task(
            task_name="servernoop", mark_pending=True
        )
        self.assertIsNone(self.playground.schedule_for_worker(worker))

    def test_get_provided_worker_tags(self) -> None:
        worker = Worker.objects.get_or_create_celery()
        system_tags = {wtags.WORKER_TYPE_PREFIX + worker.worker_type}
        for worker_tags, admin_tags, expected in (
            ([], [], set()),
            (None, None, set()),
            (["a", "b"], [], {"a", "b"}),
            ([], ["a", "b"], {"a", "b"}),
            (["a"], ["b"], {"a", "b"}),
        ):
            with self.subTest(
                worker_tags=repr(worker_tags), admin_tags=repr(admin_tags)
            ):
                worker.dynamic_metadata = {"provided_tags": worker_tags}
                worker.static_metadata = {"provided_tags": admin_tags}
                self.assertEqual(
                    get_provided_worker_tags(worker), expected | system_tags
                )

    def test_get_provided_worker_tags_worker_type(self) -> None:
        self.worker.set_dynamic_metadata(
            {
                **self.worker.dynamic_metadata,
                "provided_tags": [wtags.WORKER_TYPE_SIGNING],
            }
        )
        # Worker type uses the system version, not the worker-provided one
        self.assertEqual(
            get_provided_worker_tags(self.worker), {wtags.WORKER_TYPE_EXTERNAL}
        )


class TagBasedScheduleForWorkerTests(ScheduleForWorkerTests):
    """Version of ScheduleForWorkerTests with tag based scheduling on."""

    @override
    def setUp(self) -> None:
        super().setUp()
        self.enterContext(override_settings(FEATURE_TAG_BASED_SCHEDULING=True))

    @override_settings(DISABLE_AUTOMATIC_SCHEDULING=True)
    def test_task_required_tags_enforced(self) -> None:  # pragma: no cover
        work_request = self.create_sample_sbuild_work_request()
        for required, provided, accepted in (
            ([], [], True),
            ([], ["a"], True),
            (["a"], [], False),
            (["a"], ["a"], True),
            (["a"], ["b"], False),
            (["a", "b", "c"], ["a", "b"], False),
            (["a", "b", "c"], ["a", "b", "c"], True),
            (["a", "b", "c"], ["a", "b", "c", "d"], True),
        ):
            with (
                self.subTest(required=repr(required), provided=repr(provided)),
                mock.patch(
                    "debusine.server.scheduler.get_provided_worker_tags",
                    return_value=set(provided),
                ),
                mock.patch(
                    "debusine.server.scheduler"
                    "._assign_work_request_to_worker_or_error"
                ) as assign_work_request,
            ):
                work_request.scheduler_tags_required = required
                work_request.save()
                self.playground.schedule_for_worker(self.worker)
                if accepted:
                    assign_work_request.assert_called_with(
                        work_request, self.worker, set()
                    )
                else:
                    assign_work_request.assert_not_called()

    @override_settings(DISABLE_AUTOMATIC_SCHEDULING=True)
    def test_worker_required_tags_enforced(self) -> None:  # pragma: no cover
        # Ensure the check of worker provided tags always passes here
        work_request = self.create_sample_sbuild_work_request(
            require_scheduler_tags=False
        )
        for required, provided, accepted in (
            ([], [], True),
            ([], ["a"], True),
            (["a"], [], False),
            (["a"], ["a"], True),
            (["a"], ["b"], False),
            (["a", "b", "c"], ["a", "b"], False),
            (["a", "b", "c"], ["a", "b", "c"], True),
            (["a", "b", "c"], ["a", "b", "c", "d"], True),
        ):
            with (
                self.subTest(required=repr(required), provided=repr(provided)),
                mock.patch(
                    "debusine.server.scheduler.get_required_task_tags",
                    return_value=set(required),
                ),
                mock.patch(
                    "debusine.server.scheduler"
                    "._assign_work_request_to_worker_or_error"
                ) as assign_work_request,
            ):
                work_request.scheduler_tags_provided = provided
                work_request.save()
                assert work_request.worker is None

                self.playground.schedule_for_worker(self.worker)

                if accepted:
                    assign_work_request.assert_called_with(
                        work_request, self.worker, set()
                    )
                else:
                    assign_work_request.assert_not_called()


class ScheduleForWorkerTransactionTests(SchedulerTestCase, TransactionTestCase):
    """Test schedule_for_worker()."""

    @override
    def setUp(self) -> None:
        super().setUp()
        self.worker = self._create_worker(enabled=True)
        self.patch_field_resolver()

    @override_settings(
        DISABLE_AUTOMATIC_SCHEDULING=False, CELERY_TASK_ALWAYS_EAGER=True
    )
    async def test_schedule_from_work_request_changed(self) -> None:
        """End to end: from WorkRequest.save(), notification to schedule()."""
        work_request_1 = await database_sync_to_async(
            self.create_sample_sbuild_work_request
        )(require_scheduler_tags=False)
        self.assertIsNone(work_request_1.worker)

        # Marking work_request_1 as pending scheduled a scheduler run, which
        # assigned it to a worker.
        await work_request_1.arefresh_from_db()
        work_request_1_worker = await database_sync_to_async(
            lambda: work_request_1.worker
        )()

        self.assertEqual(work_request_1_worker, self.worker)

        # Create another work request
        work_request_2 = await database_sync_to_async(
            self.create_sample_sbuild_work_request
        )(require_scheduler_tags=False)
        self.assertIsNone(work_request_2.worker)

        await work_request_2.arefresh_from_db()
        work_request_2_worker = await database_sync_to_async(
            lambda: work_request_2.worker
        )()

        # schedule() got called but there isn't any available worker
        # (work_request_1 is assigned to self.worker and has not finished)
        self.assertIsNone(work_request_2_worker)

        # Mark work_request_1 as done transitioning to RUNNING, then
        # mark_completed(). schedule() got called when work_request_1 had
        # changed and then it assigned work_request_2 to self.worker
        await database_sync_to_async(work_request_1.mark_running)()
        await database_sync_to_async(work_request_1.mark_completed)(
            WorkRequest.Results.SUCCESS
        )

        await work_request_2.arefresh_from_db()
        work_request_2_worker = await database_sync_to_async(
            lambda: work_request_2.worker
        )()

        # Assert that work_request_2 got assigned to self.worker
        self.assertEqual(work_request_2_worker, self.worker)

    @override_settings(
        DISABLE_AUTOMATIC_SCHEDULING=False, CELERY_TASK_ALWAYS_EAGER=True
    )
    async def test_schedule_from_worker_changed(self) -> None:
        """End to end: from Worker.save() to schedule_for_worker()."""
        await database_sync_to_async(self.worker.mark_disconnected)()

        work_request_1 = await database_sync_to_async(
            self.create_sample_sbuild_work_request
        )(require_scheduler_tags=False)

        # The Worker is not connected: the work_request_1.worker is None
        self.assertIsNone(work_request_1.worker)

        # _worker_changed() will be executed via the post_save signal
        # from Worker. It will call schedule_for_worker()
        await database_sync_to_async(self.worker.mark_connected)()

        await work_request_1.arefresh_from_db()

        work_request_1_worker = await database_sync_to_async(
            lambda: work_request_1.worker
        )()

        self.assertEqual(work_request_1_worker, self.worker)

    @override_settings(DISABLE_AUTOMATIC_SCHEDULING=True)
    def test_when_worker_is_locked(self) -> None:
        """schedule_for_worker() fails when it fails to lock the Worker."""
        self.create_sample_sbuild_work_request()

        thread = RunInParallelTransaction(
            lambda: Worker.objects.select_for_update().get(id=self.worker.id)
        )
        thread.start_transaction()

        try:
            self.assertIsNone(self.playground.schedule_for_worker(self.worker))
        finally:
            thread.stop_transaction()

    @override_settings(DISABLE_AUTOMATIC_SCHEDULING=True)
    def test_when_work_request_is_locked(self) -> None:
        """schedule_for_worker() skips and logs work requests it can't lock."""
        workflow = self.playground.create_workflow()
        work_requests = [
            self.create_sample_sbuild_work_request(
                require_scheduler_tags=False, parent=parent
            )
            for parent in (workflow, workflow, None, None)
        ]

        thread = RunInParallelTransaction(
            lambda: WorkRequest.objects.filter(
                id__in={work_requests[0].id, work_requests[2].id}
            ).lock_workflows_for_update()
        )
        thread.start_transaction()

        try:
            with self.assertLogs(
                logger="debusine.server.scheduler", level=logging.DEBUG
            ) as logs:
                self.assertEqual(
                    self.playground.schedule_for_worker(self.worker),
                    work_requests[3],
                )
            # We skip work_requests[1], since we already found that another
            # work request in the same workflow was locked.
            self.assertEqual(
                logs.output,
                [
                    f"DEBUG:debusine.server.scheduler:"
                    f"[SCHEDULER] Failed to lock WorkRequest {work_request}"
                    f" (skipping this workflow)"
                    for work_request in (work_requests[0], work_requests[2])
                ],
            )
        finally:
            thread.stop_transaction()

    @override_settings(DISABLE_AUTOMATIC_SCHEDULING=True)
    def test_when_work_request_is_locked_on_error(self) -> None:
        work_request = self.create_sample_sbuild_work_request(
            require_scheduler_tags=False
        )
        # invalid: it's missing required properties
        work_request.configured_task_data = {}
        work_request.save()

        thread = RunInParallelTransaction(
            lambda: WorkRequest.objects.select_for_update().get(
                id=work_request.id
            )
        )
        thread.start_transaction()

        try:
            with self.assertLogsContains(
                f"[SCHEDULER] Failed to lock WorkRequest {work_request}"
                f" to record error",
                logger="debusine.server.scheduler",
                level=logging.DEBUG,
            ):
                self.assertIsNone(
                    self.playground.schedule_for_worker(self.worker)
                )
        finally:
            thread.stop_transaction()

    def assert_dispatches_celery_task(
        self,
        mock_apply_async: mock.MagicMock,
        work_request: WorkRequest,
        task_id_prefix: str,
    ) -> None:
        """Assert that a work request is dispatched to a Celery task."""
        mock_apply_async.assert_called_once()
        self.assertEqual(
            mock_apply_async.call_args.kwargs["args"], (work_request.id,)
        )
        self.assertRegex(
            mock_apply_async.call_args.kwargs["task_id"],
            rf"^{task_id_prefix}_{work_request.id}_"
            r"[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}$",
        )

    def find_workflow_task_results(self) -> QuerySet[TaskResult]:
        return celery_tasks_in_progress(
            "debusine.server.workflows.celery.run_workflow_task"
        ).order_by("date_created")

    @override_settings(DISABLE_AUTOMATIC_SCHEDULING=True)
    def test_server_task(self) -> None:
        """schedule_for_worker() dispatches server tasks to Celery."""
        worker = Worker.objects.get_or_create_celery()
        worker.dynamic_metadata = {
            "system:architectures": [native_architecture()],
            "system:worker_type": WorkerType.CELERY,
            "server:servernoop:version": ServerNoop.TASK_VERSION,
            "provided_tags": [
                wtags.WORKER_TYPE_CELERY,
                wtags.TASK_PREFIX
                + f"server:servernoop:version:{ServerNoop.TASK_VERSION}",
            ],
        }
        work_request = self.playground.create_server_task(
            task_name="servernoop", mark_pending=True
        )

        with mock.patch(
            "debusine.server.scheduler.run_server_task.apply_async"
        ) as mock_apply_async:
            result = self.playground.schedule_for_worker(worker)

        self.assertEqual(work_request, result)
        self.assert_dispatches_celery_task(
            mock_apply_async, work_request, "servernoop"
        )

    def test_workflow_callbacks(self) -> None:
        """schedule() dispatches workflow callbacks to Celery."""
        WorkRequest.objects.all().delete()
        parent = self.playground.create_workflow()
        parent.mark_running()
        wr = WorkRequest.objects.create_workflow_callback(
            parent=parent, step="test"
        )
        wr.mark_pending()

        with mock.patch(
            "debusine.server.scheduler.run_workflow_task.apply_async"
        ) as mock_apply_async:
            self.assertEqual(schedule(), [wr])

        self.assert_dispatches_celery_task(
            mock_apply_async, wr, "internal_workflow"
        )
        wr.refresh_from_db()
        self.assertEqual(wr.status, WorkRequest.Statuses.RUNNING)

    def test_workflow_callback_another_workflow_running(self) -> None:
        """Callbacks for multiple workflows may run concurrently."""
        test_app = TestApp(
            set_as_current=True,
            broker="memory://",
            task_cls=TaskWithBetterArgumentRepresentation,
        )
        self.enterContext(setup_default_app(test_app))

        WorkRequest.objects.all().delete()
        workflows = [self.playground.create_workflow() for _ in range(3)]
        for workflow in workflows:
            workflow.mark_running()
        wrs = [
            WorkRequest.objects.create_workflow_callback(
                parent=workflow, step="test"
            )
            for workflow in workflows[:2]
        ]
        for wr in wrs:
            wr.mark_pending()

        self.assertEqual(schedule(), wrs)

        self.assertQuerySetEqual(
            self.find_workflow_task_results()
            .filter(status=states.PENDING)
            .values_list("task_args_array", flat=True),
            [[wr.id] for wr in wrs[:2]],
        )
        self.assertQuerySetEqual(
            WorkRequest.objects.filter(id__in={wr.id for wr in wrs})
            .order_by("id")
            .values_list("status", flat=True),
            [WorkRequest.Statuses.RUNNING] * len(wrs),
        )

        wrs.append(
            WorkRequest.objects.create_workflow_callback(
                parent=workflows[-1], step="test"
            )
        )
        wrs[-1].mark_pending()
        # A running work request in the same parent workflow that isn't a
        # workflow callback doesn't stop the workflow callback running.
        self.playground.create_worker_task(
            parent=workflows[-1], assign_new_worker=True, mark_running=True
        )

        self.assertCountEqual(schedule(), [wrs[-1]])
        self.assertQuerySetEqual(
            self.find_workflow_task_results()
            .filter(status=states.PENDING)
            .values_list("task_args_array", flat=True),
            [[wr.id] for wr in wrs],
        )
        self.assertQuerySetEqual(
            WorkRequest.objects.filter(id__in={wr.id for wr in wrs})
            .order_by("id")
            .values_list("status", flat=True),
            [WorkRequest.Statuses.RUNNING] * len(wrs),
        )

    def test_workflow_callback_same_workflow_running(self) -> None:
        """Callbacks for the same workflow may not run concurrently."""
        test_app = TestApp(
            set_as_current=True,
            broker="memory://",
            task_cls=TaskWithBetterArgumentRepresentation,
        )
        self.enterContext(setup_default_app(test_app))

        WorkRequest.objects.all().delete()
        workflow = self.playground.create_workflow()
        workflow.mark_running()
        wrs = [
            WorkRequest.objects.create_workflow_callback(
                parent=workflow, step=f"test{i}"
            )
            for i in range(2)
        ]
        for wr in wrs:
            wr.mark_pending()

        self.assertEqual(schedule(), [wrs[0]])

        self.assertQuerySetEqual(
            self.find_workflow_task_results()
            .filter(status=states.PENDING)
            .values_list("task_args_array", flat=True),
            [[wrs[0].id]],
        )
        self.assertQuerySetEqual(
            WorkRequest.objects.filter(id__in={wr.id for wr in wrs})
            .order_by("id")
            .values_list("status", flat=True),
            [WorkRequest.Statuses.RUNNING, WorkRequest.Statuses.PENDING],
        )

        wrs.append(
            WorkRequest.objects.create_workflow_callback(
                parent=workflow, step="test2"
            )
        )
        wrs[-1].mark_pending()

        self.assertEqual(schedule(), [])

        workflow_task_results = self.find_workflow_task_results().filter(
            status=states.PENDING
        )
        self.assertQuerySetEqual(
            workflow_task_results.values_list("task_args_array", flat=True),
            [[wrs[0].id]],
        )
        self.assertQuerySetEqual(
            WorkRequest.objects.filter(id__in={wr.id for wr in wrs})
            .order_by("id")
            .values_list("status", flat=True),
            [
                WorkRequest.Statuses.RUNNING,
                WorkRequest.Statuses.PENDING,
                WorkRequest.Statuses.PENDING,
            ],
        )

        # Completing the Celery task (albeit artificially) allows the
        # scheduler to proceed.
        WorkRequest.objects.get(id=wrs[0].id).mark_completed(
            WorkRequest.Results.SUCCESS
        )
        workflow_task_result = workflow_task_results.get()
        workflow_task_result.status = states.SUCCESS
        workflow_task_result.save()

        self.assertEqual(schedule(), [wrs[1]])

        self.assertQuerySetEqual(
            self.find_workflow_task_results()
            .filter(status=states.PENDING)
            .values_list("task_args_array", flat=True),
            [[wrs[1].id]],
        )
        self.assertQuerySetEqual(
            WorkRequest.objects.filter(id__in={wr.id for wr in wrs})
            .order_by("id")
            .values_list("status", flat=True),
            [
                WorkRequest.Statuses.COMPLETED,
                WorkRequest.Statuses.RUNNING,
                WorkRequest.Statuses.PENDING,
            ],
        )

    def test_workflows(self) -> None:
        """schedule() dispatches workflows to Celery."""
        WorkRequest.objects.all().delete()
        wr = self.playground.create_workflow(task_name="noop")

        with mock.patch(
            "debusine.server.scheduler.run_workflow_task.apply_async"
        ) as mock_apply_async:
            self.assertEqual(schedule(), [wr])

        self.assert_dispatches_celery_task(
            mock_apply_async, wr, "workflow_noop"
        )
        wr.refresh_from_db()
        self.assertEqual(wr.status, WorkRequest.Statuses.RUNNING)

    def test_sub_workflow(self) -> None:
        """schedule() orchestrates pending sub-workflows via their root."""
        root_wr = self.playground.create_workflow(task_name="noop")
        sub_wr = self.playground.create_workflow(parent=root_wr)
        # Mark the root workflow running, to simulate the situation where it
        # can only do part of its work and has to be called back later.
        root_wr.mark_running()

        with mock.patch(
            "debusine.server.scheduler.run_workflow_task.apply_async",
            side_effect=lambda *args, **kwargs: sub_wr.mark_running(),
        ) as mock_apply_async:
            schedule()

        self.assert_dispatches_celery_task(
            mock_apply_async, root_wr, "workflow_noop"
        )
        # The (fake) orchestrator was called and marked the sub-workflow as
        # running.
        self.assertEqual(sub_wr.status, WorkRequest.Statuses.RUNNING)

    def test_groups_workflows_by_root(self) -> None:
        """schedule() groups pending workflows by their root workflow."""
        root_wr = self.playground.create_workflow(task_name="noop")
        sub_wr = self.playground.create_workflow(parent=root_wr)

        with mock.patch(
            "debusine.server.scheduler.run_workflow_task.apply_async"
        ) as mock_apply_async:
            schedule()

        # Only the root workflow's orchestrator is called.
        self.assert_dispatches_celery_task(
            mock_apply_async, root_wr, "workflow_noop"
        )
        root_wr.refresh_from_db()
        sub_wr.refresh_from_db()
        self.assertEqual(root_wr.status, WorkRequest.Statuses.RUNNING)
        self.assertEqual(sub_wr.status, WorkRequest.Statuses.PENDING)

    def test_sub_workflow_another_workflow_running(self) -> None:
        """Sub-workflows of different root workflows may run concurrently."""
        test_app = TestApp(
            set_as_current=True,
            broker="memory://",
            task_cls=TaskWithBetterArgumentRepresentation,
        )
        self.enterContext(setup_default_app(test_app))

        WorkRequest.objects.all().delete()
        roots = [self.playground.create_workflow() for _ in range(3)]
        # Mark the root workflows running, to simulate the situation where
        # they can only do part of their work and have to be called back
        # later.
        for root in roots:
            root.mark_running()
        for root in roots[:2]:
            child = self.playground.create_workflow(parent=root)
            child.mark_running()
            self.playground.create_workflow(parent=child)

        self.assertEqual(schedule(), roots[:2])

        self.assertQuerySetEqual(
            self.find_workflow_task_results()
            .filter(status=states.PENDING)
            .values_list("task_args_array", flat=True),
            [[root.id] for root in roots[:2]],
        )

        self.playground.create_workflow(parent=roots[-1])
        # A running work request in the same parent workflow that isn't a
        # sub-workflow doesn't stop the sub-workflow running.
        self.playground.create_worker_task(
            parent=roots[-1], assign_new_worker=True, mark_running=True
        )

        self.assertCountEqual(schedule(), [roots[-1]])

        self.assertQuerySetEqual(
            self.find_workflow_task_results()
            .filter(status=states.PENDING)
            .values_list("task_args_array", flat=True),
            [[root.id] for root in roots],
        )

    def test_sub_workflow_same_workflow_running(self) -> None:
        """Sub-workflows of the same root workflow may not run concurrently."""
        test_app = TestApp(
            set_as_current=True,
            broker="memory://",
            task_cls=TaskWithBetterArgumentRepresentation,
        )
        self.enterContext(setup_default_app(test_app))

        WorkRequest.objects.all().delete()
        root = self.playground.create_workflow()
        root.mark_running()
        for _ in range(2):
            self.playground.create_workflow(parent=root)

        self.assertEqual(schedule(), [root])

        self.assertQuerySetEqual(
            self.find_workflow_task_results()
            .filter(status=states.PENDING)
            .values_list("task_args_array", flat=True),
            [[root.id]],
        )

        self.playground.create_workflow(parent=root)

        self.assertEqual(schedule(), [])

        workflow_task_results = self.find_workflow_task_results().filter(
            status=states.PENDING
        )
        self.assertQuerySetEqual(
            workflow_task_results.values_list("task_args_array", flat=True),
            [[root.id]],
        )

        # Completing the Celery task (albeit artificially) allows the
        # scheduler to proceed.
        workflow_task_result = workflow_task_results.get()
        workflow_task_result.status = states.SUCCESS
        workflow_task_result.save()

        self.assertEqual(schedule(), [root])

        self.assertQuerySetEqual(
            self.find_workflow_task_results()
            .filter(status=states.PENDING)
            .values_list("task_args_array", flat=True),
            [[root.id]],
        )


class UpdateWorkflowTransactionTests(TransactionTestCase):
    """Unit tests for :py:func:`update_workflows`."""

    @override_settings(PGTRANSACTION_RETRY=0)
    def test_updates_workflows(self) -> None:
        with transaction.atomic():
            workflow = self.playground.create_workflow()
            work_request = workflow.create_child_worker("noop")
            self.assertFalse(work_request.workflows_need_update)
            work_request.mark_aborted()
            self.assertTrue(work_request.workflows_need_update)
            unrelated = self.playground.create_worker_task()

        with transaction.atomic():
            schedule()

            workflow.refresh_from_db()
            work_request.refresh_from_db()
            self.assertEqual(
                workflow.workflow_last_activity_at, work_request.completed_at
            )
            self.assertEqual(
                workflow.workflow_runtime_status,
                WorkRequest.RuntimeStatuses.ABORTED,
            )
            self.assertFalse(work_request.workflows_need_update)
            self.assert_work_requests_locked(
                locked=[workflow, work_request], unlocked=[unrelated]
            )

    def test_repeatable_read(self) -> None:
        """The function runs in the correct transaction isolation level."""
        with (
            transaction.atomic(),
            mock.patch("django.db.transaction.on_commit"),
        ):
            workflows = [self.playground.create_workflow() for _ in range(2)]
            work_requests = [
                workflow.create_child_worker("noop") for workflow in workflows
            ]
            self.assertFalse(work_requests[0].workflows_need_update)
            work_requests[0].mark_aborted()
            self.assertTrue(work_requests[0].workflows_need_update)
            self.assertFalse(work_requests[1].workflows_need_update)

        def create_workflow_needs_update() -> None:
            with transaction.atomic():
                work_requests[1].mark_aborted()
                self.assertTrue(work_requests[1].workflows_need_update)

        def create_workflow_needs_update_in_thread(
            *args: Any, **kwargs: Any
        ) -> Any:
            thread = RunInParallelTransaction(create_workflow_needs_update)
            thread.run_and_wait()
            return mock.DEFAULT

        with mock.patch(
            "debusine.server.scheduler.workflow_ancestors",
            side_effect=create_workflow_needs_update_in_thread,
            wraps=workflow_ancestors,
        ):
            schedule()

        # The first workflow is updated.
        workflows[0].refresh_from_db()
        work_requests[0].refresh_from_db()
        self.assertEqual(
            workflows[0].workflow_last_activity_at,
            work_requests[0].completed_at,
        )
        self.assertFalse(work_requests[0].workflows_need_update)

        # The second workflow is untouched, since it didn't need an update
        # until after our function's transaction started.
        workflows[1].refresh_from_db()
        work_requests[1].refresh_from_db()
        self.assertIsNone(workflows[1].workflow_last_activity_at)
        self.assertTrue(work_requests[1].workflows_need_update)

    def test_concurrent_updates(self) -> None:
        """
        If the workflow lock query hits a serialization failure, the
        function aborts with an ``OperationalError``.
        """
        with transaction.atomic():
            workflow = self.playground.create_workflow()
            work_request = workflow.create_child_worker("noop")
            work_request.mark_aborted()

        with (
            self.assertLogsContains(
                "update_workflows() deferred due to a concurrent update",
                logger="debusine.server.scheduler",
                level=logging.WARNING,
            ),
            mock.patch(
                "debusine.db.models.work_requests.WorkRequestQuerySet.lock_workflows_for_update",
                side_effect=OperationalError(
                    "could not serialize access due to concurrent update"
                ),
            ),
        ):
            schedule()


class ScheduleSoonTests(TestCase):
    """Test schedule_soon()."""

    @override_settings(CELERY_TASK_ALWAYS_EAGER=True, PGTRANSACTION_RETRY=0)
    def test_runs(self) -> None:
        with (
            mock.patch("debusine.server.scheduler.schedule") as mock_schedule,
            self.captureOnCommitCallbacks(execute=True),
        ):
            schedule_soon()

        mock_schedule.assert_called_once()

    def test_batches(self) -> None:
        """schedule_soon() runs the scheduler once per savepoint."""
        with (
            transaction.atomic(),
            mock.patch(
                "debusine.server.scheduler.schedule_task.apply_async"
            ) as mock_apply_async,
            self.captureOnCommitCallbacks(execute=True) as callbacks,
        ):
            schedule_soon()
            schedule_soon()

        self.assertEqual(
            len(
                [
                    task
                    for task in get_apply_async_tasks(callbacks)
                    if task == schedule_task
                ]
            ),
            1,
        )
        mock_apply_async.assert_called_once()



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Notifications using channels or NotificationChannels."""

import logging
import string
from typing import Any, TYPE_CHECKING

from asgiref.sync import async_to_sync
from channels import DEFAULT_CHANNEL_LAYER
from channels.layers import channel_layers
from channels_redis.core import RedisChannelLayer
from django.core.mail import EmailMessage

from debusine.tasks.models import ActionSendNotification, EventReaction

if TYPE_CHECKING:
    from debusine.db.models import NotificationChannel, Token, WorkRequest

logger = logging.getLogger(__name__)


async def _group_send_and_close(key: str, message: dict[str, Any]) -> None:
    """
    Send a message to a group via a channel layer, then close pools if needed.

    See https://github.com/django/channels_redis/issues/332.
    """
    channel_layer = channel_layers[DEFAULT_CHANNEL_LAYER]
    await channel_layer.group_send(key, message)
    if isinstance(channel_layer, RedisChannelLayer):  # pragma: no cover
        await channel_layer.close_pools()


def notify_work_request_assigned(work_request: "WorkRequest") -> None:
    """Send a channel message of type 'work_request_assigned'."""
    if work_request.worker is None or work_request.worker.token is None:
        return

    async_to_sync(_group_send_and_close)(
        work_request.worker.token.hash, {"type": "work_request.assigned"}
    )


def notify_worker_token_disabled(token: "Token") -> None:
    """Notify to the WorkerConsumer that a worker has been disabled."""
    from debusine.db.models import Token

    if token.token_type != Token.Types.WORKER:
        return

    async_to_sync(_group_send_and_close)(
        token.hash, {"type": "worker.disabled"}
    )


def notify_work_request_completed(
    work_request: "WorkRequest", notifications: list[EventReaction]
) -> None:
    """Notify to the interested parties that work_request is completed."""
    # Send to the channel_layer (Websockets) that the work request has completed
    assert work_request.completed_at is not None
    async_to_sync(_group_send_and_close)(
        "work_request_completed",
        {
            "type": "work_request_completed",
            "work_request_id": work_request.id,
            "workspace_id": work_request.workspace.id,
            "completed_at": work_request.completed_at.isoformat(),
            "result": work_request.result,
        },
    )

    from debusine.db.models import NotificationChannel

    # Prepare the notifications to the Notification channel based on the
    # WorkRequest "notifications" section in event_reactions (e.g. email
    # notifications)
    for work_request_notification in notifications:
        assert isinstance(work_request_notification, ActionSendNotification)
        name = work_request_notification.channel
        try:
            notification_channel = NotificationChannel.objects.get(name=name)
        except NotificationChannel.DoesNotExist:
            logger.debug(
                "WorkRequest %s: work-request-completed notification "
                "cannot be sent: "
                "NotificationChannel %s does not exist",
                work_request.id,
                name,
            )
            continue

        body = "Sent by debusine"

        _send_work_request_completed_notification(
            notification_channel,
            work_request,
            work_request_notification,
            body,
        )


def _send_work_request_completed_notification(
    notification_channel: "NotificationChannel",
    work_request: "WorkRequest",
    work_request_notification: ActionSendNotification,
    body: str,
) -> None:
    from debusine.db.models import NotificationChannel

    if notification_channel.method != NotificationChannel.Methods.EMAIL:
        raise NotImplementedError()

    email_data = {**notification_channel.data}
    if work_request_notification.data is not None:
        email_data.update(
            (k, v) for k, v in work_request_notification.data if v is not None
        )

    if (subject := email_data.get("subject")) is None:
        subject = (
            "WorkRequest $work_request_id completed in $work_request_result"
        )
    subject = string.Template(subject).safe_substitute(
        work_request_id=work_request.id, work_request_result=work_request.result
    )

    email = EmailMessage(
        subject=subject,
        body=body,
        from_email=email_data["from"],
        to=email_data["to"],
        cc=email_data.get("cc"),
    )

    success = email.send(fail_silently=True)

    if success == 0:
        logger.warning(
            "NotificationChannel email notification for WorkRequest "
            "%s failed sending",
            work_request.id,
        )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Signing task to sign a repository index."""

import shutil
from pathlib import Path
from typing import override

from debusine.artifacts import RepositoryIndex
from debusine.artifacts.models import ArtifactCategory, WorkRequestResults
from debusine.assets import AssetCategory, KeyPurpose
from debusine.client.models import AssetPermissionCheckResponse, RelationType
from debusine.signing.db.models import Key, sign
from debusine.signing.models import SigningMode
from debusine.signing.tasks import BaseSigningTask
from debusine.signing.tasks.models import (
    SignRepositoryIndexData,
    SignRepositoryIndexDynamicData,
)
from debusine.tasks.inputs import SingleInput, SuiteArchiveInput


class SignRepositoryIndex(
    BaseSigningTask[SignRepositoryIndexData, SignRepositoryIndexDynamicData]
):
    """Task that signs a repository index using a key."""

    TASK_VERSION = 1

    unsigned = SingleInput(categories=[ArtifactCategory.REPOSITORY_INDEX])
    suite_archive = SuiteArchiveInput(field="suite_collection")

    @override
    def post_init(self) -> None:
        super().post_init()
        self._file_path: str | None = None
        self._keys: list[Key] = []
        self._permission_check: AssetPermissionCheckResponse | None = None
        self._signed_path: Path | None = None

    @override
    def get_subject(self) -> str | None:
        return None

    @override
    def compute_dynamic_data(self) -> SignRepositoryIndexDynamicData:
        """Resolve lookups for this task."""
        suite, archive = self.suite_archive
        signing_keys: list[str] | None = None
        if suite.data.get("signing_keys") is not None:
            signing_keys = suite.data["signing_keys"]
        else:
            if archive is not None:
                signing_keys = archive.data.get("signing_keys")

        return SignRepositoryIndexDynamicData(
            signing_keys=signing_keys or [],
            unsigned_id=self.unsigned.artifact_id,
        )

    @override
    def fetch_input(self, destination: Path) -> bool:
        """Fetch artifacts and database objects needed by the task."""
        assert self.debusine is not None
        assert self.dynamic_data is not None
        assert self.work_request_id is not None
        assert self.workspace_name is not None

        if not self.dynamic_data.signing_keys:
            self.append_to_log_file(
                "fetch_input.log",
                [
                    f"No signing keys are configured for "
                    f"{self.data.suite_collection}."
                ],
            )
            return False

        unsigned = self.fetch_artifact(
            self.dynamic_data.unsigned_id, destination
        )
        # debian:repository-index artifacts always contain exactly one file.
        assert len(unsigned.files) == 1
        [self._file_path] = unsigned.files.keys()

        self._keys = []
        for key in self.dynamic_data.signing_keys:
            asset_slug = f"{KeyPurpose.OPENPGP_REPOSITORY}:{key}"
            permission_check = self.debusine.asset_permission_check(
                asset_category=AssetCategory.SIGNING_KEY,
                asset_slug=asset_slug,
                permission_name="sign_repository_with",
                artifact_id=self.dynamic_data.unsigned_id,
                work_request_id=self.work_request_id,
                workspace=self.workspace_name,
            )
            if not permission_check.has_permission:
                self.append_to_log_file(
                    "fetch_input.log",
                    [
                        f"User {permission_check.username} does not have "
                        f"permission to sign {self.dynamic_data.unsigned_id} "
                        f"with {asset_slug}."
                    ],
                )
                return False
            if (
                self._permission_check is not None
                and permission_check != self._permission_check
            ):
                self.append_to_log_file(
                    "fetch_input.log",
                    [
                        f"Mismatching permission-check responses for "
                        f"different signing keys: "
                        f"{permission_check} != {self._permission_check}"
                    ],
                )
                return False
            self._permission_check = permission_check

            try:
                self._keys.append(
                    Key.objects.get(
                        purpose=KeyPurpose.OPENPGP_REPOSITORY, fingerprint=key
                    )
                )
            except Key.DoesNotExist:
                self.append_to_log_file(
                    "fetch_input.log",
                    [f"Signing key {asset_slug} does not exist"],
                )
                return False

        return True

    @override
    def configure_for_execution(
        self,
        download_directory: Path,  # noqa: U100
    ) -> bool:
        """Configure task variables."""
        # Nothing to do here; fetch_input does it all.
        return True

    @override
    def prepare_to_run(
        self, download_directory: Path, execute_directory: Path
    ) -> None:
        """Copy downloaded files into the execution directory."""
        shutil.copytree(download_directory, execute_directory / "input")

    @override
    def run(self, execute_directory: Path) -> WorkRequestResults:
        """Sign the requested file."""
        assert self.work_request_id is not None
        assert self._file_path is not None
        assert self._permission_check is not None

        with self.open_debug_log_file("cmd-output.log", mode="wb") as log_file:
            data_path = execute_directory / "input" / self._file_path
            signed_path = execute_directory / "output" / self.data.signed_name
            signed_path.parent.mkdir(parents=True, exist_ok=True)
            try:
                sign(
                    self._keys,
                    data_path,
                    signed_path,
                    self.data.mode,
                    self.work_request_id,
                    username=self._permission_check.username,
                    user_id=self._permission_check.user_id,
                    resource=self._permission_check.resource,
                    log_file=log_file,
                )
            except Exception as e:
                log_file.write(f"Signing failed: {e}\n".encode())
                raise
            else:
                self._signed_path = signed_path

        return WorkRequestResults.SUCCESS

    @override
    def upload_artifacts(
        self,
        execute_directory: Path,
        *,
        execution_result: WorkRequestResults,
    ) -> None:
        """Upload artifacts for the task."""
        assert self.dynamic_data is not None
        assert self.work_request_id is not None
        assert self.debusine is not None
        assert self._signed_path is not None
        # run() always either returns SUCCESS or raises an exception; in the
        # latter case we won't get here.
        assert execution_result == WorkRequestResults.SUCCESS

        content_type: str | None
        match self.data.mode:
            case SigningMode.DETACHED:
                content_type = "application/pgp-signature; charset=utf-8"
            case SigningMode.CLEAR:
                # This is unregistered in
                # https://www.iana.org/assignments/media-types/media-types.xhtml,
                # but it's what file(1) claims.
                content_type = "text/PGP; charset=utf-8"
            case _:  # pragma: no cover
                # Checked by SignRepositoryIndexData.validate_mode.
                raise AssertionError(f"Unexpected mode: {self.data.mode}")

        signed_artifact = RepositoryIndex.create(
            file=self._signed_path,
            path=self.data.signed_name,
            content_type=content_type,
        )
        uploaded_signed_artifact = self.debusine.upload_artifact(
            signed_artifact,
            workspace=self.workspace_name,
            work_request=self.work_request_id,
        )
        self.debusine.relation_create(
            uploaded_signed_artifact.id,
            self.dynamic_data.unsigned_id,
            RelationType.EXTENDS,
        )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""debusine-admin command to manage file stores."""

import argparse
import sys
from collections.abc import Callable
from typing import Any, IO, NoReturn, cast, override

from django.core.exceptions import ValidationError
from django.core.management import CommandError, CommandParser
from django.db import IntegrityError, transaction

from debusine.assets.models import AssetCategory
from debusine.db.models import Asset, FileStore
from debusine.django.management.debusine_base_command import DebusineBaseCommand


class Command(DebusineBaseCommand):
    """Command to manage file stores."""

    help = "Manage file stores."

    @override
    def add_arguments(self, parser: CommandParser) -> None:
        subparsers = parser.add_subparsers(dest="action", required=True)

        create = subparsers.add_parser(
            "create", help="Ensure a file store exists"
        )
        create.add_argument("name", help="Name")
        create.add_argument(
            "backend",
            help="Backend type",
            choices=dict(FileStore.BackendChoices.choices).keys(),
        )
        create.add_argument(
            "--configuration",
            type=argparse.FileType("r"),
            help=(
                "File path (or - for stdin) to read the configuration for the "
                "file store. YAML format. Defaults to stdin."
            ),
            default="-",
        )
        create.add_argument(
            "--instance-wide",
            action=argparse.BooleanOptionalAction,
            help=(
                "Whether this file store may be used by any scope "
                "(default: --instance-wide)"
            ),
        )
        create.add_argument(
            "--soft-max-size",
            type=int,
            help=(
                "Soft limit in bytes for the total capacity of the store; may "
                "be exceeded temporarily during uploads, and cleaned up later "
                "by storage maintenance"
            ),
        )
        create.add_argument(
            "--max-size",
            type=int,
            help=(
                "Hard limit in bytes for the total capacity of the store; may "
                "not be exceeded even temporarily during uploads"
            ),
        )
        create.add_argument(
            "--provider-account",
            type=self.get_provider_account,
            metavar="name",
            help="Name of cloud provider account",
        )

        delete = subparsers.add_parser("delete", help="Delete a file store")
        delete.add_argument("name", help="Name")
        delete.add_argument(
            "--force",
            action="store_true",
            help=(
                "Delete the file store even if it still contains files or is "
                "still in any scope"
            ),
        )

    @override
    def cleanup_arguments(self, *args: Any, **options: Any) -> None:
        if "configuration" in options and options["configuration"] != sys.stdin:
            options["configuration"].close()

    def get_provider_account(self, name: str) -> Asset:
        """Look up a cloud provider account asset by name."""
        try:
            return Asset.objects.get(
                category=AssetCategory.CLOUD_PROVIDER_ACCOUNT, data__name=name
            )
        except Asset.DoesNotExist:
            raise CommandError(
                f"Cloud provider account asset {name!r} not found", returncode=3
            )

    def handle_create(
        self,
        *,
        name: str,
        backend: str,
        configuration: IO[Any],
        instance_wide: bool | None = None,
        soft_max_size: int | None = None,
        max_size: int | None = None,
        provider_account: Asset | None = None,
        **options: Any,
    ) -> NoReturn:
        """
        Ensure a file store exists.

        This is idempotent, to make it easier to invoke from Ansible.
        """
        configuration_data = self.parse_yaml_data(configuration.read()) or {}

        with transaction.atomic():
            file_store, _ = FileStore.objects.get_or_create(
                name=name,
                defaults={
                    "backend": backend,
                    "configuration": configuration_data,
                    "instance_wide": (
                        True if instance_wide is None else instance_wide
                    ),
                    "soft_max_size": soft_max_size,
                    "max_size": max_size,
                    "provider_account": provider_account,
                },
            )
            if file_store.backend != backend:
                raise CommandError(
                    "Cannot change backend of existing file store", returncode=3
                )
            if file_store.configuration != configuration_data:
                # In some cases this might be possible, but in general it is
                # unsafe: for example, changing the base directory of a
                # local file store would orphan any existing files there.
                raise CommandError(
                    "Cannot change configuration of existing file store",
                    returncode=3,
                )
            if (
                instance_wide is not None
                and file_store.instance_wide != instance_wide
            ):
                file_store.instance_wide = instance_wide
            if (
                soft_max_size is not None
                and file_store.soft_max_size != soft_max_size
            ):
                file_store.soft_max_size = soft_max_size
            if max_size is not None and file_store.max_size != max_size:
                file_store.max_size = max_size
            if (
                provider_account is not None
                and provider_account != file_store.provider_account
            ):
                file_store.provider_account = provider_account

            try:
                file_store.full_clean()
                file_store.save()
            except ValidationError as exc:
                raise CommandError(
                    "Error creating file store: " + "\n".join(exc.messages),
                    returncode=3,
                )
            except IntegrityError as exc:
                raise CommandError(
                    f"Error creating file store: {exc}", returncode=3
                )

        raise SystemExit(0)

    def handle_delete(
        self, *, name: str, force: bool, **options: Any
    ) -> NoReturn:
        """Delete the file store."""
        with transaction.atomic():
            try:
                file_store = FileStore.objects.get(name=name)
            except FileStore.DoesNotExist:
                raise CommandError(
                    f"File store {name!r} does not exist", returncode=3
                )

            if (num_files := file_store.files.count()) and not force:
                raise CommandError(
                    f"File store {name!r} still contains {num_files} files; "
                    f"drain it first",
                    returncode=3,
                )

            if (num_scopes := file_store.scopes.count()) and not force:
                raise CommandError(
                    (
                        f"File store {name!r} is still used by {num_scopes} "
                        f"scopes; use 'debusine-admin scope remove_file_store' "
                        f"first"
                    ),
                    returncode=3,
                )

            file_store.files.clear()
            file_store.scopes.clear()
            file_store.delete()

        raise SystemExit(0)

    @override
    def handle(self, *args: Any, **options: Any) -> NoReturn:
        """Dispatch the requested action."""
        func = cast(
            Callable[..., NoReturn],
            getattr(self, f"handle_{options['action']}", None),
        )
        if func is None:
            raise CommandError(
                f"Action {options['action']!r} not found", returncode=3
            )

        func(*args, **options)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Data models for the db application."""

from debusine.db.models.artifacts import (
    Artifact,
    ArtifactRelation,
    FileInArtifact,
    FileUpload,
)
from debusine.db.models.assets import Asset, AssetUsage
from debusine.db.models.auth import (
    Group,
    Identity,
    SYSTEM_USER_NAME,
    Token,
    User,
    system_user,
)
from debusine.db.models.collections import (
    Collection,
    CollectionItem,
    CollectionItemMatchConstraint,
    CollectionRelation,
    CollectionTokenGrant,
)
from debusine.db.models.files import (
    DEFAULT_FILE_STORE_NAME,
    File,
    FileInStore,
    FileStore,
    default_file_store,
)
from debusine.db.models.scopes import FileStoreInScope, Scope
from debusine.db.models.work_requests import (
    NotificationChannel,
    WellKnownWorkflowTemplate,
    WorkRequest,
    WorkflowTemplate,
)
from debusine.db.models.worker_pools import (
    ScopeWorkerPool,
    WorkerPool,
    WorkerPoolStatistics,
    WorkerPoolTaskExecutionStatistics,
)
from debusine.db.models.workers import Worker
from debusine.db.models.workspaces import Workspace, default_workspace

__all__ = [
    "DEFAULT_FILE_STORE_NAME",
    "SYSTEM_USER_NAME",
    "Artifact",
    "ArtifactRelation",
    "Asset",
    "AssetUsage",
    "Collection",
    "CollectionItem",
    "CollectionItemMatchConstraint",
    "CollectionRelation",
    "CollectionTokenGrant",
    "File",
    "FileInArtifact",
    "FileInStore",
    "FileStore",
    "FileStoreInScope",
    "FileUpload",
    "Group",
    "Identity",
    "NotificationChannel",
    "Scope",
    "ScopeWorkerPool",
    "Token",
    "User",
    "WellKnownWorkflowTemplate",
    "WorkRequest",
    "Worker",
    "WorkerPool",
    "WorkerPoolStatistics",
    "WorkerPoolTaskExecutionStatistics",
    "WorkflowTemplate",
    "Workspace",
    "default_file_store",
    "default_workspace",
    "system_user",
]
